from infi.vendata.vmware_powertools.utils import get_infinidat_device_by_serial
from infi.vendata.vmware_powertools.utils import get_infinisdk_host_and_volume_by_device
import logging

logger = logging.getLogger(__name__)


class NoHostsException(Exception):
    pass


def is_cluster_mapped(infinidat_device):
    """
    Returns whether the given device is mapped to a cluster (and not to the specific host)
    """
    host, volume = get_infinisdk_host_and_volume_by_device(infinidat_device)
    if host is None or (volume is None or volume in [mapping.volume for mapping in host.get_luns()
                                                     if not mapping.clustered]):
        return False
    return True

def is_host_connected(host):
    return host.runtime.connectionState == "connected" and \
           host.runtime.powerState == "poweredOn"

def is_host_in_maintenance_mode(host):
    return host.runtime.inMaintenanceMode

def sort_hosts_by_name(hosts):
    return sorted(hosts, key=lambda host: host.name)

def good_cluster_hosts(cluster, maintenance_hosts=True):
    if maintenance_hosts:
        result = [h for h in cluster.host if is_host_connected(h)]
    else:
        result = [h for h in cluster.host if is_host_connected(h) and not is_host_in_maintenance_mode(h)]
    if len(result) == 0:
        raise NoHostsException("No operational ESX hosts found in cluster to perform requested operation")
    return result

def get_all_host_mappings(volume_serial):
    """
    Returns the morefs of hosts that the given volume is mapped to (including cluster-mapped hosts).
    """
    from infi.vendata.vmware_powertools.vcenter import vCenter
    client = vCenter.get_extension_client()
    host_mappings = []
    for esx_host in sort_hosts_by_name(filter(is_host_connected, client.get_host_systems())):
        device = get_infinidat_device_by_serial(esx_host, volume_serial)
        if device:
            host_mappings.append(esx_host)
    return sort_hosts_by_name(host_mappings)

def get_all_mappings_in_cluster(esx_cluster, volume_serial):
    """
    Returns the morefs of hosts in the cluster that the given volume is mapped to.
    """
    host_mappings = set()
    for esx_host in good_cluster_hosts(esx_cluster, maintenance_hosts=False):
        device = get_infinidat_device_by_serial(esx_host, volume_serial)
        if device:
            host_mappings.add(esx_host)
    return sort_hosts_by_name(host_mappings)
