from os import path, makedirs
from infi.pyutils.lazy import cached_function
from logging import getLogger
from .. import DATA_DIR

logger = getLogger(__name__)

# http://antirez.com/post/redis-persistence-demystified.html
# http://redis.io/topics/persistence

REDIS_CONFIGURATION = {
                       'port': 6380,
                       'bind': '127.0.0.1',
                       'timeout': 0,
                       'loglevel': 'notice',
                       'logfile': path.join(DATA_DIR, 'redis.log'),
                       'databases': 1,
                       'dbfilename': 'redis.rdb',
                       'dir': DATA_DIR,
                       'appendonly': 'yes',
                       'appendfilename': 'redis.aof',
                       'appendfsync': 'always'
                       }

REDIS_CONF = path.join(DATA_DIR, 'redis.conf')

class RedisServer(object):
    _pid = None

    @classmethod
    def _execute_redis_server(cls):
        from infi.execute import execute_async
        return execute_async(['redis-server', REDIS_CONF])

    @classmethod
    def _wait_for_redis(cls):
        from redis import ConnectionError
        client = cls.get_client()
        try:
            client.get('this')
        except ConnectionError:  # pragma: no cover
            cls._wait_for_redis()

    @classmethod
    def _write_configuration_file(cls):
        if not path.exists(DATA_DIR):
            makedirs(DATA_DIR)  # pragma: no cover
        with open(REDIS_CONF, 'w') as fd:
            fd.write("\n".join(["{} {}".format(key, value) for key, value in REDIS_CONFIGURATION.items()]))

    @classmethod
    def start(cls):
        from atexit import register
        cls._write_configuration_file()
        cls._pid = cls._execute_redis_server()
        cls._wait_for_redis()
        logger.info("Redis server started")
        register(cls.stop)

    @classmethod
    def is_running(cls):
        return cls._pid is not None and not cls._pid.is_finished()

    @classmethod
    def stop(cls):
        if not cls.is_running():  # pragma: no cover
            return
        cls._pid.kill()
        logger.info("Redis server stopped")

    @classmethod
    @cached_function
    def get_client(cls):
        import redis
        pool = redis.ConnectionPool(host=REDIS_CONFIGURATION['bind'], port=REDIS_CONFIGURATION['port'], db=0)
        return redis.Redis(connection_pool=pool)

    @classmethod
    def clear_cache(cls):
        from infi.pyutils.lazy import clear_cache
        clear_cache(cls.get_client)
