from infi.vendata.vmware_powertools.vcenter import vCenter
from infi.vendata.vmware_powertools.vcenter.shortcuts import object_to_moref, wait_for_task
from .vendor_query import VendorQuery
from .snapshotstore import SnapshotStore
from . import exceptions
import uuid
import logging

logger = logging.getLogger(__name__)


CONSISTENCY_SNAPSHOT_NAME = "INFINIDAT Consistency Snapshot"


class CreateSnapshotOperation(object):
    def __init__(self, requesting_object, vm_list, snapshot_name, snapshot_description, consistent, consolidate, snapshot_object_name=None):
        import time
        from gevent.lock import Semaphore as Lock

        if not snapshot_object_name:
            self._snapshot_object_name = "snap_vm_{}".format(uuid.uuid4().hex)
        self._requesting_object = requesting_object
        self._vm_list = vm_list
        self._snapshot_name = snapshot_name
        self._snapshot_description = snapshot_description
        self._consistent = consistent
        self._consolidate = consolidate
        self._snapshot_object_name = snapshot_object_name
        self._snapshot_time = time.localtime()
        self._lock = Lock()

    def _get_vm_folder(self, vm):
        parent_folder = vm.parent
        if parent_folder is not None:
            return parent_folder
        vapp = vm.parentVApp
        while vapp is not None:
            if vapp.parentFolder is not None:
                return vapp.parentFolder
            vapp = vapp.parentVApp
        # vm.parent is None and couldn't find parentVApp that has a non-null parentFolder. deep stuff.
        # if you ever get here, that means we there's a case we didn't think of. Maybe try to find the
        # the datacenter's vmFolder from vm.summary.runtime.host.<parent.parent...>.vmFolder
        raise RuntimeError("Could not find parent folder for VM '{}'".format(vm.name))

    def _prepare_snapshot_info(self, vm, datastore, vendor_data, infinisdk_snapshot, snap_group_uuid):
        import time
        vm_datastore_path = vm.config.files.vmPathName.replace(datastore.summary.name, "{}", 1)
        snapshot_time = time.strftime("%Y/%m/%d %H:%M:%S", self._snapshot_time)
        parent_folder = self._get_vm_folder(vm)
        return dict(snapshot_name=self._snapshot_name,
                    snapshot_description=self._snapshot_description,
                    snapshot_time=snapshot_time,
                    infinibox_serial=vendor_data.get_system_serial(),
                    infinibox_snapshot_id=infinisdk_snapshot.get_id(),
                    vm_datastore_path=vm_datastore_path,
                    snapshot_id=uuid.uuid4().hex,
                    original_folder_ref=object_to_moref(parent_folder),
                    original_host_uuid=vm.summary.runtime.host.summary.hardware.uuid,
                    original_datastore_url=datastore.info.url,
                    vm_uuid=vm.summary.config.uuid,
                    vm_name=vm.name,
                    snap_group_uuid=snap_group_uuid,
                    consistency_snapshot=self._consistent)

    def _take_vm_snapshots(self, vm_list, snap_group_uuid):
        create_snapshot_tasks = []
        result = []
        for vm in vm_list:
            create_task = vm.CreateSnapshot_Task(name=CONSISTENCY_SNAPSHOT_NAME,
                description=snap_group_uuid, memory=False, quiesce=True)
            create_snapshot_tasks.append(create_task)
        for task in create_snapshot_tasks:
            wait_for_task(task)
            result.append(task.info.result)
        logger.debug("created {} VM snapshots".format(len(result)))
        return result

    def _delete_vm_snapshots(self, created_snapshots):
        logger.debug("deleting {} VM snapshots".format(len(created_snapshots)))
        delete_snapshot_tasks = []
        for created_snapshot in created_snapshots:
            task = created_snapshot.RemoveSnapshot_Task(False, self._consolidate)
            delete_snapshot_tasks.append(task)
        for task in delete_snapshot_tasks:
            wait_for_task(task)

    def _create_infinidat_snapshot(self, vendor_data):
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device_vendor
        infinisdk_system = get_system_object_by_device_vendor(vendor_data)
        infinisdk_volume = infinisdk_system.volumes.get(id=vendor_data.get_volume_id())
        infinisdk_snapshot = infinisdk_volume.create_snapshot(self._snapshot_object_name)
        return infinisdk_snapshot

    def _create_datastore_snapshot(self, datastore_vms, snap_group_uuid):
        """
        Create snapshots for all virtual machines in datastore_vms, assuming all are on the same datastore so only one
        volume snapshot needs to be taken
        """
        from infi.vendata.vmware_powertools.patches import with_host

        datastore = VendorQuery.vm_datastores(datastore_vms[0])[0]
        if not all(all(vm_datastore.summary.url == datastore.summary.url
                       for vm_datastore in VendorQuery.vm_datastores(vm))
                   for vm in datastore_vms):
            raise exceptions.NotInfinidatDatastore("Not all Virtual Machines are on the same datastore")

        vendor_data = VendorQuery.get_datastore_infinidat_vendor(datastore)
        if vendor_data is None:
            raise exceptions.NotInfinidatDatastore("Not INFINIDAT datastore")

        if not datastore.host:
            raise exceptions.HostNotFound("Cannot find any hosts in the datastore")
        host = datastore.host[0].key

        msg = "creating datastore snapshots of {} vms. datastore url={}, snap_group_uuid={}"
        logger.debug(msg.format(len(datastore_vms), datastore.info.url, snap_group_uuid))

        created_snapshots = []
        try:
            if self._consistent:
                created_snapshots.extend(self._take_vm_snapshots(datastore_vms, snap_group_uuid))
            with with_host(host):
                infinisdk_snapshot = self._create_infinidat_snapshot(vendor_data)
        finally:
            self._delete_vm_snapshots(created_snapshots)

        metadata = dict(name=self._snapshot_name,
                        snap_group_uuid=snap_group_uuid,
                        first_vm_uuid=datastore_vms[0].summary.config.uuid,
                        vm_refs=",".join([object_to_moref(vm) for vm in datastore_vms]),
                        description=self._snapshot_description,
                        vm_list=",".join([vm.name for vm in datastore_vms]),
                        original_datastore=datastore.name)
        infinisdk_snapshot.set_metadata_from_dict(metadata)

        self._lock.acquire()
        try:
            snap_group_list = SnapshotStore.get_group_list(snap_group_uuid)

            for vm in datastore_vms:
                snapshot_info = self._prepare_snapshot_info(vm, datastore, vendor_data, infinisdk_snapshot,
                                                            snap_group_uuid)
                if infinisdk_snapshot.get_metadata_value('first_vm_snapshot_id', None) is None:
                    infinisdk_snapshot.set_metadata_from_dict({'first_vm_snapshot_id': snapshot_info['snapshot_id']})
                    infinisdk_snapshot.refresh()
                for store, key in [(SnapshotStore.VM_SNAP_STORE, snapshot_info['vm_uuid']),
                                   (SnapshotStore.FOLDER_SNAP_STORE, snapshot_info['original_folder_ref']),
                                   (SnapshotStore.DATASTORE_SNAP_STORE, snapshot_info['original_datastore_url'])]:
                    SnapshotStore.add_to_store(store, key, snapshot_info)
                    snap_group_list.append((store, key, snapshot_info['snapshot_id']))

            SnapshotStore.set_group_list(snap_group_uuid, snap_group_list)
        finally:
            self._lock.release()

    def create(self):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
        from infi.vendata.vmware_powertools.utils import spawn, joinall
        logger.debug("creating snapshot of %d vms" % len(self._vm_list))
        self._client = vCenter.get_extension_client()
        self._task = create_vcenter_task(self._requesting_object, "Create INFINIDAT Snapshot")

        with self._task:
            if not self._snapshot_name:
                raise exceptions.SnapshotException('Snapshot name cannot be empty')
            snap_group_uuid = uuid.uuid4().hex
            # divide vm_list to datastores
            datastore_dict = dict()
            for vm in self._vm_list:
                if len(VendorQuery.vm_datastores(vm)) != 1:
                    raise exceptions.NotInfinidatDatastore("Found virtual machine that has more than one "
                                                           "datastore attached.")
                datastore_url = VendorQuery.vm_datastores(vm)[0].summary.url
                datastore_dict.setdefault(datastore_url, []).append(vm)
            joinall([spawn(self._create_datastore_snapshot, datastore_vms, snap_group_uuid)
                     for datastore_vms in datastore_dict.values()], raise_error=True)
