from infi.pyutils.contexts import contextmanager
from infi.vendata.vmware_powertools.vcenter import vCenter
from infi.vendata.vmware_powertools.vcenter.shortcuts import wait_for_task
from infi.vendata.vmware_powertools.utils import spawn, joinall
from infi.vendata.vmware_powertools.mapping import good_cluster_hosts
from infi.vendata.vmware_powertools.vcenter.shortcuts import get_cluster_datastores
from . import exceptions
from .snapshotstore import SnapshotStore
from .create import CONSISTENCY_SNAPSHOT_NAME
from pyVmomi import vim
import uuid
import logging

logger = logging.getLogger(__name__)


def find_new_name(client, snapshot_info):
    VSPHERE_MAX_CHARS = 80
    SUFFIX_LENGTH = len("-revert000")
    original_vm_name = snapshot_info["vm_name"][:(VSPHERE_MAX_CHARS - SUFFIX_LENGTH)]
    count = 0
    while count < 1000:
        new_name = "{}-revert{}".format(original_vm_name, "" if count == 0 else count)
        if client.get_virtual_machine(new_name) is None:
            return new_name
        count += 1
    # if we got here that means that all 1000 snapshot names are taken!
    logger.info("Couldn't find new name for reverted VM. User will have to choose...")
    return ""


def get_default_revert_params(client, vm, snapshot_info):
    from infi.vendata.vmware_powertools.vcenter.shortcuts import get_host_by_uuid, get_datastore_by_url
    logger.debug("original host uuid is {}. original datastore url is {}. Folder moref is {}".format(
        snapshot_info["original_host_uuid"],
        snapshot_info["original_datastore_url"],
        snapshot_info["original_folder_ref"]))
    host = get_host_by_uuid(client, snapshot_info["original_host_uuid"])
    if host is None:
        if vm is not None:
            host = vm.summary.runtime.host
            logger.info("Using vm's current host, uuid {}".format(host.summary.hardware.uuid))
        else:
            logger.info("Original host not found. User will have to choose...")
    datastore = get_datastore_by_url(client, snapshot_info["original_datastore_url"])
    if datastore is None:
        logger.info("Original datastore not found. Trying to choose datastore from host")
        if host is not None:
            datastore = host.datastore[0]
        else:
            logger.info("User will have to choose datatore...")
    try:
        folder = client.get_managed_object_by_reference(snapshot_info["original_folder_ref"])
    except ValueError:
        logger.info("Original folder not found. Using host's datacenter's main folder")
        if host is not None:
            datacenter = host.parent.parent.parent
            folder = datacenter.vmFolder
        else:
            folder = None
            logger.info("User will have to choose folder...")
    name = find_new_name(client, snapshot_info)
    logger.info("New suggested name is {}".format(name))
    return name, host, datastore, folder


def wait_for_volume(host, volume):
    from infi.vendata.vmware_powertools.patches import with_host
    with with_host(host):
        from infi.storagemodel.vendor.infinidat.predicates import InfinidatVolumeExists
        from infi.storagemodel import get_storage_model
        model = get_storage_model()
        predicate = InfinidatVolumeExists(volume.get_system().get_serial(), volume.get_id())
        model.rescan_and_wait_for(predicate)


def find_volume_device(host, volume):
    from infi.vendata.vmware_powertools.patches import with_host
    with with_host(host):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
        for dev in get_infinidat_block_devices():
            vendor = dev.get_vendor()
            volume_id = volume.get_id()
            system_serial = volume.get_system().get_serial()
            if vendor.get_volume_id() == volume_id \
               and vendor.get_system_serial() == system_serial:
                return dev
    raise exceptions.VolumeNotMapped("Could not find device mapped for volume id {}".format(volume.get_id()))


def mount_datastore_snapshot(host, volume_device):
    device_path = volume_device.get_block_access_path() + ":1"
    resignature_spec = vim.HostUnresolvedVmfsResignatureSpec(extentDevicePath=device_path)
    task = host.configManager.datastoreSystem.ResignatureUnresolvedVmfsVolume_Task(resignature_spec)
    wait_for_task(task)
    result = task.info.result.result
    return result


class RevertSnapshotOperation(object):
    def __init__(self, requesting_object, vm_uuid, snapshot_id, resignature_vm, vm_name, vm_obj, datastore, folder):
        self._requesting_object = requesting_object
        self._should_resignature_vm = resignature_vm
        self._snapshot_info = SnapshotStore.get_snapshot_info(SnapshotStore.VM_SNAP_STORE, vm_uuid, snapshot_id)
        self._update_call_count = 0
        self._vm_name = vm_name
        self._vm_obj = vm_obj
        if isinstance(vm_obj, vim.ClusterComputeResource):
            self._host = good_cluster_hosts(vm_obj, maintenance_hosts=False)[0]
            self._resource_pool = vm_obj.resourcePool
            self._use_resource_pool = True
            self._use_cluster_mapping = True
        elif isinstance(vm_obj, vim.HostSystem):
            self._host = vm_obj
            self._resource_pool = vm_obj.parent.resourcePool
            self._use_resource_pool = False
            self._use_cluster_mapping = False
        else:   # "vm_obj" is a resourcepool/vapp
            self._host = vm_obj.owner.host[0]
            self._resource_pool = vm_obj
            self._use_resource_pool = True
            self._use_cluster_mapping = False
        self._datastore = datastore
        self._folder = folder

    @contextmanager
    def _map_snapshot(self, clean=True):
        from infi.vendata.vmware_powertools.patches import with_host
        from infi.vendata.vmware_powertools.utils import get_system_object_and_host, get_system_object_and_cluster
        from infi.vendata.powertools.utils.queries import ObjectNotFound
        vm_obj = self._vm_obj if self._use_cluster_mapping else self._host
        vm_obj_type = "host" if not self._use_cluster_mapping else "cluster"
        self._update_task_description("Mapping snapshot volume to {} {}".format(vm_obj_type, vm_obj.name))
        infinibox_serial = self._snapshot_info["infinibox_serial"]
        try:
            with with_host(self._host):
                if not self._use_cluster_mapping:
                    infinisdk_system, infinisdk_object = get_system_object_and_host(infinibox_serial)
                else:
                    infinisdk_system, infinisdk_object = get_system_object_and_cluster(infinibox_serial)
        except ObjectNotFound:
            msg = "Selected {} '{}' is not registered in system {}."
            raise ObjectNotFound(msg.format(vm_obj_type, vm_obj.name, infinibox_serial))
        infinibox_snapshot_id = self._snapshot_info['infinibox_snapshot_id']
        logger.debug("Infinibox Snapshot id={}".format(infinibox_snapshot_id))
        self._snapshot_vol = infinisdk_system.volumes.safe_get(id=infinibox_snapshot_id)
        if self._snapshot_vol is None:
            raise exceptions.SnapshotNotFound("Can't find snapshot object on InfiniBox {}".format(infinibox_serial))
        cloned_vol = self._snapshot_vol.create_child()
        cloned_vol.disable_write_protection()
        infinisdk_object.map_volume(cloned_vol)
        try:
            yield cloned_vol
        except:
            # expose operations will not clean the volume in the 'finally' clause - so for exceptions on expose,
            # we need to clean explicitly here
            if not clean:
                infinisdk_object.unmap_volume(cloned_vol)
                cloned_vol.delete()
            # we don't rescan because we don't know what caused the exception -
            # it may be because the VC is down for example
        finally:
            if clean:
                self._update_task_description("Cleaning volume")
                infinisdk_object.unmap_volume(cloned_vol)
                cloned_vol.delete()
                self._update_task_description("Rescanning {}".format(vm_obj_type))
                if not self._use_cluster_mapping:
                    self._rescan_host(vm_obj)
                else:
                    def op(host):
                        self._rescan_host(host)
                    joinall([spawn(op, host) for host in vm_obj.host], raise_error=True)

    def _rescan_host(self, host):
        from infi.vendata.vmware_powertools.patches import with_host
        with with_host(host):
            from infi.storagemodel import get_storage_model
            model = get_storage_model()
            model.rescan_and_wait_for()

    @contextmanager
    def _find_new_datastore(self, cloned_vol, clean=True):
        self._update_task_description("Rescanning host")
        wait_for_volume(self._host, cloned_vol)
        self._update_task_description("Locating device")
        volume_device = find_volume_device(self._host, cloned_vol)
        self._update_task_description("Resolving new VMFS volume")
        new_datastore = mount_datastore_snapshot(self._host, volume_device)
        if self._use_cluster_mapping:
            # rescaning the other hosts will also auto-mount the newly resignatured datastore on them
            self._update_task_description("Rescanning other hosts in cluster")
            def op(other_host):
                wait_for_volume(other_host, cloned_vol)
            other_hosts = good_cluster_hosts(self._vm_obj, maintenance_hosts=False)[1:]
            joinall([spawn(op, other_host) for other_host in other_hosts], raise_error=True)
        try:
            yield new_datastore
        finally:
            if clean:
                self._update_task_description("Cleaning snapshot datastore")
                self._host.configManager.storageSystem.UnmountVmfsVolume(new_datastore.info.vmfs.uuid)
                try:
                    self._host.configManager.storageSystem.DetachScsiLun(volume_device.get_uuid())
                except vim.HostConfigFault:
                    logger.error("Got HostConfigFault on DetachScsiLun. See HIPVM-601")

    def _register_vm(self, new_datastore):
        self._update_task_description("Registering Virtual Machine")
        datastore_path = self._snapshot_info["vm_datastore_path"].format(new_datastore.name)
        if not self._vm_name:
            raise exceptions.NameNotAvailable("Empty name is not allowed.")
        if self._client.get_virtual_machine(self._vm_name):
            raise exceptions.NameNotAvailable("The chosen name is already in use.")
        # for hosts and resource pools that are not children of vapps, we need to use folder.RegisterVM
        # for vapps and resource pools/vapps that are children of vapps, we need to use resource_pool.RegisterChildVM
        if self._use_resource_pool and \
           (isinstance(self._resource_pool, vim.VirtualApp) or isinstance(self._resource_pool.parent, vim.VirtualApp)):
            task = self._resource_pool.RegisterChildVM_Task(datastore_path, name=self._vm_name, host=None)
        else:
            host = None if self._use_resource_pool else self._host
            task = self._folder.RegisterVM_Task(datastore_path,
                                                name=self._vm_name,
                                                asTemplate=False,
                                                pool=self._resource_pool,
                                                host=host)
        wait_for_task(task)
        return task.info.result

    def _relocate_vm(self, vm, datastore):
        import time
        self._update_task_description("Relocating Virtual Machine")
        relocate_spec = vim.VirtualMachineRelocateSpec(datastore=datastore,
                                                       transform=None,
                                                       host=None,
                                                       pool=None)
        task = vm.RelocateVM_Task(relocate_spec, None)
        try:
            wait_for_task(task)
        except vim.NoDiskSpace as e:
            # the following reraise will cause a popup in the browser with something like:
            # "ERROR: Insufficient disk space on datastore 'v7'."
            raise exceptions.NoDiskSpace(e.msg)
        # For some weird reason, we can't delete the snapshot immediately after relocation, so sleep a bit...
        time.sleep(5)

    def _update_task_description(self, new_description):
        logger.debug("Task description: {}".format(new_description))
        self._task.set_description(new_description)
        percentage = int((self._update_call_count) * (100.0 / self._operation_count))
        self._update_call_count += 1
        self._task.update_progress(percentage)

    def _resignature_vm(self, vm):
        self._update_task_description("Resignaturing Virtual Machine")
        cfg_spec = vim.VirtualMachineConfigSpec(uuid=str(uuid.uuid1()))
        task = vm.ReconfigVM_Task(cfg_spec)
        wait_for_task(task)

    def _unregister_vm(self, vm):
        vm.UnregisterVM()

    def _get_snapshots_by_name(self, vm, name):
        from collections import deque
        if vm.snapshot:
            queue = deque(vm.snapshot.rootSnapshotList)
            while queue:
                node = queue.popleft()
                if node.name == name:
                    yield node
                queue.extend(node.childSnapshotList)

    def _revert_to_consistency_snapshot(self, vm):
        matches = list(self._get_snapshots_by_name(vm, CONSISTENCY_SNAPSHOT_NAME))
        # there may be consistency snapshots from previous snaps+reverts.
        # the snapshot description is the uuid of the snap group, it must match
        matches = [match for match in matches if match.description == self._snapshot_info["snap_group_uuid"]]
        if matches:
            task = matches[-1].snapshot.RevertToSnapshot_Task()
            self._client.wait_for_task(task)

    def _clone_snapshots_to_restored_vm(self, new_vm):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import object_to_moref
        original_vm_uuid = self._snapshot_info['vm_uuid']
        new_vm_uuid = new_vm.summary.config.uuid
        snapshot_list = SnapshotStore.get_from_store(SnapshotStore.VM_SNAP_STORE, original_vm_uuid)
        for snapshot_info in snapshot_list:
            new_snapshot_info = snapshot_info
            new_snapshot_info.update(dict(snapshot_id=uuid.uuid4().hex,
                                          vm_uuid=new_vm_uuid,
                                          vm_name=new_vm.name))
            SnapshotStore.add_to_store(SnapshotStore.VM_SNAP_STORE, new_vm_uuid, new_snapshot_info)
            group_uuid = snapshot_info['snap_group_uuid']
            group_list = SnapshotStore.get_group_list(group_uuid)
            group_list.append((SnapshotStore.VM_SNAP_STORE, new_vm_uuid, new_snapshot_info['snapshot_id']))
            SnapshotStore.set_group_list(group_uuid, group_list)
        self._snapshot_vol.set_metadata_from_dict(dict(
            vm_refs=self._snapshot_vol.get_metadata_value("vm_refs", "") + "," + object_to_moref(new_vm),
            vm_list=self._snapshot_vol.get_metadata_value("vm_list", "") + "," + new_vm.name))

    def revert(self, expose_only=False):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
        logger.debug(("exposing" if expose_only else "restoring") + " snapshot")
        self._client = vCenter.get_extension_client()
        task_name = ("Expose" if expose_only else "Restore") + " INFINIDAT Snapshot"
        self._task = create_vcenter_task(self._requesting_object, task_name)
        self._operation_count = 5           # map, rescan, locate, resolve, register
        if self._use_cluster_mapping:
            self._operation_count += 1      # rescan cluster
        if not expose_only:
            self._operation_count += 4      # relocate, clean ds, clean vol, rescan
        if self._should_resignature_vm:
            self._operation_count += 1      # resig

        with self._task:

            # make sure datastore is visible to host
            logger.debug("uuid of host to be used is {}".format(self._host.summary.hardware.uuid))
            if not expose_only:
                # datastore is not important if we expose_only, but if we need to relocate, then
                # the datastore details are important and the check is needed
                if not self._use_resource_pool:
                    all_datastores = self._host.datastore
                else:
                    all_datastores = get_cluster_datastores(self._resource_pool.owner)
                logger.debug("all datastores in target: {}".format(all_datastores))
                logger.debug("datastore to relocate to: {}".format(self._datastore))
                if self._datastore not in all_datastores:
                    raise exceptions.DatastoreNotAvailable("Target datastore is not visible to target host or cluster")

            with self._map_snapshot(clean=not expose_only) as cloned_vol:
                with self._find_new_datastore(cloned_vol, clean=not expose_only) as new_datastore:
                    new_vm = self._register_vm(new_datastore)
                    if not expose_only:
                        try:
                            self._relocate_vm(new_vm, self._datastore)
                        except:
                            self._unregister_vm(new_vm)
                            raise
                    if self._should_resignature_vm:
                        self._resignature_vm(new_vm)
                        self._clone_snapshots_to_restored_vm(new_vm)
                    self._revert_to_consistency_snapshot(new_vm)

        if expose_only:
            return self._vm_name, new_datastore.name, cloned_vol.get_name()
        else:
            return self._vm_name, None, None
