from infi.vendata.vmware_powertools.vcenter import vCenter
from ..patches import ContextQueue
from .snapshotstore import SnapshotStore
from .create import CreateSnapshotOperation
from .revert import RevertSnapshotOperation
from .register import RegisterSnapshotOperation
from . import exceptions


def get_infinisdk_snapshot_volume(snapshot_info):
    from infi.vendata.vmware_powertools.vcenter.shortcuts import get_host_by_uuid
    from infi.vendata.vmware_powertools.patches import with_host
    from infi.vendata.vmware_powertools.utils import get_system_object_by_serial
    client = vCenter.get_extension_client()
    original_host = get_host_by_uuid(client, snapshot_info["original_host_uuid"])
    infinibox_snapshot_id = snapshot_info['infinibox_snapshot_id']
    infinibox_serial = snapshot_info["infinibox_serial"]
    with with_host(original_host):
        infinisdk_system = get_system_object_by_serial(infinibox_serial)
    snapshot = infinisdk_system.volumes.safe_get(id=infinibox_snapshot_id)
    if snapshot is None:
        raise exceptions.SnapshotNotFound("Can't find snapshot object on InfiniBox {}".format(infinibox_serial))
    return snapshot


class SnapshotOperations(object):
    _revert_queue = ContextQueue()

    @classmethod
    def create_snapshot(cls, requesting_object, vm_list, snapshot_name, snapshot_description, consistent, consolidate, snapshot_object_name=None):
        CreateSnapshotOperation(requesting_object, vm_list, snapshot_name, snapshot_description, consistent, consolidate, snapshot_object_name).create()

    @classmethod
    def register_snapshot(cls, requesting_object, vm_list, snapshot_id):
        RegisterSnapshotOperation(requesting_object, vm_list, snapshot_id).register()

    @classmethod
    def revert_snapshot(cls, requesting_object, vm_uuid, snapshot_id, resignature_vm, vm_name, host, datastore, folder, expose_only=False):
        with cls._revert_queue.wait_for_your_turn():
            return RevertSnapshotOperation(requesting_object, vm_uuid, snapshot_id, resignature_vm, vm_name, host, datastore, folder).revert(expose_only=expose_only)

    @classmethod
    def remove_snapshot(cls, snapshot_info):
        snap_group_uuid = snapshot_info["snap_group_uuid"]
        snap_group_list = SnapshotStore.get_group_list(snap_group_uuid)
        deleted_volume_ids = set()
        for store, key, snapshot_id in snap_group_list:
            cur_snapshot_info = SnapshotStore.get_snapshot_info(store, key, snapshot_id)
            snap_vol_identifier = (cur_snapshot_info["infinibox_serial"], cur_snapshot_info["infinibox_snapshot_id"])
            if snap_vol_identifier not in deleted_volume_ids:
                try:
                    infinisdk_snapshot = get_infinisdk_snapshot_volume(cur_snapshot_info)
                    infinisdk_snapshot.delete()
                except exceptions.SnapshotNotFound:
                    pass
                deleted_volume_ids.add(snap_vol_identifier)
            SnapshotStore.del_from_store(store, key, snapshot_id)

        SnapshotStore.del_group_list(snap_group_uuid)

    @classmethod
    def update_snapshot(cls, snapshot_info):
        new_name = snapshot_info["snapshot_name"]
        new_description = snapshot_info["snapshot_description"]

        snap_group_list = SnapshotStore.get_group_list(snapshot_info["snap_group_uuid"])

        updated_volume_ids = []

        for store, key, snapshot_id in snap_group_list:
            cur_snapshot_info = SnapshotStore.get_snapshot_info(store, key, snapshot_id)
            cur_snapshot_info["snapshot_name"] = new_name
            cur_snapshot_info["snapshot_description"] = new_description
            SnapshotStore.add_to_store(store, key, cur_snapshot_info)

            snap_vol_identifier = (cur_snapshot_info["infinibox_serial"], cur_snapshot_info["infinibox_snapshot_id"])
            if snap_vol_identifier not in updated_volume_ids:
                infinisdk_snapshot = get_infinisdk_snapshot_volume(cur_snapshot_info)
                infinisdk_snapshot.set_metadata_from_dict(dict(name=new_name, description=new_description))
                updated_volume_ids.append(snap_vol_identifier)
