import time
from infi.monotonic_time import monotonic_time
from ZSI import FaultException
from .zsi.vasaService_types import ns2
from .utils import client_sessions, decorators
from .entities import Entities
from logging import getLogger
logger = getLogger(__name__)

# we must not return more than 100 alarms in a single request. described in the VASA spec
MAX_ALARMS = 100
# how much time until next check (vCenter requests alarms every 10 seconds but we don't want to check so frequently)
CHECK_INTERVAL_IN_SECONDS = 60

class VasaProviderAlarmsMixin(object):
    def __init__(self):
        # dict of session-id -> next alarm id for the session
        self.session_to_alarm_id = dict()
        # dict of session-id -> alarms currently active in sessions
        self.session_to_alarms = dict()
        # to avoid too many requests, we only check after a certain time. This is used for time tracking
        self.last_check_time = None

    def _init_alarm_session(self, session_id):
        self.session_to_alarm_id[session_id] = 0
        self.session_to_alarms[session_id] = list()

    def _deinit_alarm_session(self, session_id):
        self.session_to_alarm_id.pop(session_id, None)
        self.session_to_alarms.pop(session_id, None)

    def _next_alarm_id(self):
        session_id = client_sessions.get_session().get_session_id()
        next_id = self.session_to_alarm_id.setdefault(session_id, 0)
        self.session_to_alarm_id[session_id] += 1
        return next_id

    def _generate_alarm(self, object_id, status):
        alarm_id = self._next_alarm_id()
        result = self.local.response.new_return()
        result.set_element_alarmId(alarm_id)
        result.set_element_alarmType("SpaceCapacity")
        result.set_element_alarmTimeStamp(time.localtime())
        result.set_element_objectId(object_id)
        result.set_element_objectType("StorageLun")
        result.set_element_status(status)
        result.set_element_messageId("ThinProvisioningAlarm.message")
        result.set_element_parameterList([])
        return result

    def _get_real_status_of_device(self, volume):
        pool = volume.get_pool()
        if pool.is_over_critical_threshold():
            return "Red"
        if pool.is_over_warning_threshold():
            return "Yellow"
        return None

    def _get_current_status_of_device(self, session_alarms, object_id):
        current_alarms_for_device = [alarm for alarm in session_alarms
                                     if alarm.get_element_objectId() == object_id]
        if len(current_alarms_for_device) == 0:
            return None
        return current_alarms_for_device[0].get_element_status()

    def _check_alarm_for_device(self, device, session_alarms):
        object_id = Entities.unique_lun_id(device)
        volume = self._volume_from_device(device)
        if volume is None:
            logger.warn("Could not get volume for device {}".format(object_id))
            return
        if volume.get_provisioning() != "THIN":
            return
        if object_id not in self.ids_ever_returned:
            logger.warn("device has not been returned from VASA yet. Skipping. {}".format(object_id))
            return
        # found a thin provisioned volume
        # figure out "current status" (alarms in session_alarms that were already sent to vCenter)
        current_status = self._get_current_status_of_device(session_alarms, object_id)
        # figure out "real status" according to actual state
        real_status = self._get_real_status_of_device(volume)
        # decide what to do to sync real_status and current_status
        if current_status == real_status:
            # synced. Either no current alarm and no real alarm, or the alarm was already sent.
            return
        if current_status is not None:
            # an alarm exists in session_alarms that should be removed or replaced. remove alarm from session_alarms
            logger.debug("removing alarm for {}".format(object_id))
            # the [:] is used to filter in-place, otherwise a new list will be created and
            # self.session_to_alarms won't be modified
            session_alarms[:] = [alarm for alarm in session_alarms if alarm.get_element_objectId() != object_id]
            if real_status is None:
                real_status = "Green"       # went from bad to good, so generate green alarm
        if real_status is not None:
            # add new alarm to session_alarms
            logger.debug("generating new alarm for {} with status {}".format(object_id, real_status))
            alarm = self._generate_alarm(object_id, real_status)
            session_alarms.append(alarm)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_getAlarms(self):
        last_id = self.local.request.get_element_lastAlarmId()
        if last_id < -1:
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())

        # get "session_alarms" which will be used throughout the flow. This list contains the alarms
        # that were previously sent, to determine if an alarm is new or not. New alrams are added to this
        # list and we return the alarms from this list.
        session_id = client_sessions.get_session().get_session_id()
        session_alarms = self.session_to_alarms.setdefault(session_id, list())

        # time check to avoid heavy load
        if self.last_check_time is not None and monotonic_time() < self.last_check_time + CHECK_INTERVAL_IN_SECONDS:
            # less than CHECK_INTERVAL_IN_SECONDS passed since last traversal. skip the check this time
            # (but still return results from session_alarms, if alarms already exist with id after last_id)
            logger.debug("skipping getAlarms check (too frequent)")
        else:
            for device in Entities.get_device_iterator():
                self._check_alarm_for_device(device, session_alarms)
            self.last_check_time = monotonic_time()

        result = [alarm for alarm in session_alarms if alarm.get_element_alarmId() > last_id]
        logger.debug("getAlarms: {} alarms to return".format(len(result)))
        self.local.response.set_element_return(result[:MAX_ALARMS])
