from itertools_recipes import unique_everseen
from .utils import client_sessions
from infi.pyutils.lazy import cached_function, clear_cache
from infi.vendata.vmware_powertools.utils.auth import auth

from logging import getLogger
logger = getLogger(__name__)

ARRAY_ID_PREFIX = "Infinibox-"

def iterate_hosts_with_callback(callback):
    """ calls "callback" inside a host context, for each host in the vCenter, and
    aggregates items from the callback calls """
    from infi.vendata.vmware_powertools.patches import with_host
    from infi.vendata.vmware_powertools.vcenter import vCenter
    host_initiator = client_sessions.get_session().get_session_value('host_initiator')
    if not host_initiator:
        # techinically, we should only iterate over hosts that are in the host_initiator list from setContext
        # but we'll just go over all hosts in the vCenter we are connected to instead (we will not support
        # two vCenters with different systems attached to each). However, to pass the certification, we will
        # not go over any host if the host initiator list is empty or not present
        return
    client = vCenter.get_extension_client()
    esx_hosts = client.get_host_systems()
    for esx_host in esx_hosts:
        with with_host(esx_host):
            for item in callback():
                yield item


class Entities(object):
    """ Implements iterators and unique ids for every entity supported by VASA:
    1. arrays (systems), 2. processors (nodes), 3. ports, 4. LUNs (volumes/devices), 5. capabilities """

    @classmethod
    def get_controller_iterator(cls):
        def callback():
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
            return get_infinidat_block_devices_and_controllers__mapped_to_lun0()
        return unique_everseen(iterate_hosts_with_callback(callback),
                               key=lambda controller: cls.unique_array_id(controller))

    @classmethod
    def get_device_iterator(cls):
        def callback():
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
            return get_infinidat_block_devices()
        return unique_everseen(iterate_hosts_with_callback(callback),
                               key=lambda device: cls.unique_lun_id(device))

    @classmethod
    def get_port_iterator(cls):
        from infi.storagemodel.connectivity import FCConnectivity
        ports = []
        for controller in cls.get_controller_iterator():
            ports.extend([path.get_connectivity()._remote_port for path in controller.get_paths()
                          if isinstance(path.get_connectivity(), FCConnectivity)])
        return unique_everseen(ports, key=lambda port: port.port_wwn)

    @classmethod
    def _test_system_access(cls, system):
        from infinisdk.core.exceptions import APICommandFailed
        try:
            # assure read access
            system.login()
            system.api.get('system')
            return True
        except APICommandFailed:
            return False

    @classmethod
    @cached_function
    def _find_credentials_for_system_cache(cls, system, serial):
        logger.debug("entered _find_credentials_for_system_cache. serial {}".format(serial))
        credentials_to_try = auth.CREDENTIALS_STORE.get_all_possible_credentials(serial)
        for credentials in credentials_to_try:
            logger.debug("trying existing credentials for VASA use. username {}".format(credentials.get_username()))
            system.api.set_auth((credentials.get_username(), credentials.get_password()), login=False)  # login will happen in '_test_system_access'
            if cls._test_system_access(system):
                return credentials
        return None

    @classmethod
    def _find_credentials_for_system(cls, system, serial):
        from infi.credentials_store import NoCredentialsException
        credentials = cls._find_credentials_for_system_cache(system, serial)
        logger.debug("_find_credentials_for_system_cache result: serial {} credentials {}".format(serial, credentials))
        if not credentials:
            raise NoCredentialsException()
        # we found working credentials for this system, maybe from the cache
        system.api.set_auth((credentials.get_username(), credentials.get_password()), login=False)  # login will happen in '_test_system_access'
        # cached credentials may change and stop working so we test access again here, in the non-cached method
        if not cls._test_system_access(system):
            logger.info("credentials test failed! resetting")
            # reset the cache and try again
            clear_cache(cls._find_credentials_for_system_cache)
            # if we succeed in the next call, the auth will be set correctly and be cached for next time too
            if not cls._find_credentials_for_system_cache(system, serial):
                raise NoCredentialsException()

    @classmethod
    def get_system_object_by_device(cls, device):
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device
        from infi.vendata.powertools.utils.versioncheck import UnsupportedVersion
        from infi.credentials_store import NoCredentialsException
        from infinisdk.core.exceptions import APICommandFailed
        serial = device.get_vendor().get_system_serial()
        try:
            system = get_system_object_by_device(device, log_in=False)
            cls._find_credentials_for_system(system, serial)
            return system
        except UnsupportedVersion:
            logger.info("Unsupported version for system {}".format(serial))
        except APICommandFailed as e:
            if e.status_code != 401:
                raise
            logger.info("Unauthorized login for system {}".format(serial))
        except NoCredentialsException:
            logger.info("No credentials for system {}".format(serial))

    @classmethod
    def get_infinisdk_dict(cls):
        result = dict()
        for controller in cls.get_controller_iterator():
            serial = controller.get_vendor().get_system_serial()
            if serial not in result:
                infinisdk_system = cls.get_system_object_by_device(controller)
                if infinisdk_system:
                    result[infinisdk_system.get_serial()] = infinisdk_system
        return result

    @classmethod
    def get_capability_iterator(self):
        return [("VMFS;High Performance", "InfiniRaid high performance storage")]

    @classmethod
    def get_node_iterator(cls):
        for box_serial, infinisdk_system in cls.get_infinisdk_dict().items():
            for server in infinisdk_system.components.nodes.to_list():
                yield server

    @classmethod
    def unique_array_id(cls, controller):
        return "{}{}".format(ARRAY_ID_PREFIX, controller.get_vendor().get_system_serial())

    @classmethod
    def unique_node_id(cls, node):
        return "{}-Node-{}".format(node.get_system().get_serial(), node.get_id())

    @classmethod
    def unique_lun_id(cls, device):
        return str(device.get_vendor().get_naa())

    @classmethod
    def unique_port_id(cls, port):
        return str(port.port_wwn)

    @classmethod
    def unique_capability_id(cls, capability):
        return "Capability-{}".format(hash(capability))


# trick to clear the _find_credentials_for_system cache when setting/resetting credentials
def clear_cache_decorator(func):
    from infi.pyutils.decorators import wraps
    @wraps(func)
    def wrapped(*args, **kwargs):
        clear_cache(Entities._find_credentials_for_system_cache)
        return func(*args, **kwargs)
    return wrapped


auth.CREDENTIALS_STORE.set_credentials = clear_cache_decorator(auth.CREDENTIALS_STORE.set_credentials)
auth.CREDENTIALS_STORE.reset_credentials = clear_cache_decorator(auth.CREDENTIALS_STORE.reset_credentials)
