from .entities import Entities
from .utils import decorators
from logging import getLogger
logger = getLogger(__name__)

class VasaProviderEntityAssociationMixin(object):
    """ implements queryAssociatedProcessorsForArray, queryAssociatedPortsForProcessor,
    queryAssociatedLunsForPort, queryAssociatedCapabilityForLun"""

    def _create_associations(self, associations):
        """ associations is list of tuples entity_id->associated_id.
        This function sets the response with these associations, making sure
        to add only one associations per entity, with all of its associated ids, without repetition """
        # convert tuple to dict of [entity id]->[set of associated ids]
        associations_dict = dict()
        for entity_id, associated_id in associations:
            associations_dict.setdefault(entity_id, set()).add(associated_id)
        vasa_associations = []
        for entity_id, associated_id_set in associations_dict.items():
            association = self.local.response.new_return()
            entity_id_storage_item = association.new_entityId()
            entity_id_storage_item.set_element_uniqueIdentifier(entity_id)
            self.ids_ever_returned.add(entity_id)
            associaed_items_list = []
            for associated_id in associated_id_set:
                associated_id_storage_item = association.new_associatedId()
                associated_id_storage_item.set_element_uniqueIdentifier(associated_id)
                associaed_items_list.append(associated_id_storage_item)
                self.ids_ever_returned.add(associated_id)
            association.set_element_entityId([entity_id_storage_item])
            association.set_element_associatedId(associaed_items_list)
            vasa_associations.append(association)

        self.local.response.set_element_return(vasa_associations)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryAssociatedProcessorsForArray(self):
        requested_ids = self.local.request.get_element_arrayUniqueId()
        iterator = Entities.get_controller_iterator()
        id_builder = Entities.unique_array_id
        requested_controllers = self._request_ids_to_items(requested_ids, iterator, id_builder)
        associations = []
        for controller in requested_controllers:
            infinisdk_system = Entities.get_system_object_by_device(controller)
            if infinisdk_system is None:
                continue
            for server in infinisdk_system.components.nodes.to_list():
                node_id = Entities.unique_node_id(server)
                associations.append((id_builder(controller), node_id))
        self._create_associations(associations)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryAssociatedPortsForProcessor(self):
        from infi.storagemodel.connectivity import FCConnectivity
        requested_ids = self.local.request.get_element_spUniqueId()
        iterator = Entities.get_node_iterator()
        id_builder = Entities.unique_node_id
        requested_nodes = self._request_ids_to_items(requested_ids, iterator, id_builder)

        initator_to_targets = {}
        for controller in Entities.get_controller_iterator():
            connections = [path.get_connectivity() for path in controller.get_paths()
                           if isinstance(path.get_connectivity(), FCConnectivity)]
            controller_targets = {str(item.get_target_wwn()) for item in connections}
            controller_initiators = {str(item.get_initiator_wwn()) for item in connections}
            for initiator in controller_initiators:
                initator_to_targets.setdefault(initiator, set()).update(controller_targets)

        associations = []
        for node in requested_nodes:
            for target in node.get_fc_ports():
                target_id = str(target.get_wwpn())
                associations.append((id_builder(node), target_id))
        self._create_associations(associations)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryAssociatedLunsForPort(self):
        requested_ids = self.local.request.get_element_portUniqueId()
        iterator = Entities.get_port_iterator()
        id_builder = Entities.unique_port_id
        requested_ports = self._request_ids_to_items(requested_ids, iterator, id_builder)

        lun_id_to_wwns = {}
        for device in Entities.get_device_iterator():
            device_wwns = {str(path.get_connectivity().get_target_wwn()) for path in device.get_paths()}
            lun_id_to_wwns[Entities.unique_lun_id(device)] = device_wwns

        associations = []
        for port in requested_ports:
            for lun_id, wwns in lun_id_to_wwns.items():
                logger.debug("soap_queryAssociatedLunsForPort port={} wwns={}".format(id_builder(port), wwns))
                if id_builder(port) in wwns:
                    associations.append((id_builder(port), lun_id))
        self._create_associations(associations)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryAssociatedCapabilityForLun(self):
        requested_ids = self.local.request.get_element_lunUniqueId()
        iterator = Entities.get_device_iterator()
        id_builder = Entities.unique_lun_id
        requested_luns = self._request_ids_to_items(requested_ids, iterator, id_builder)
        # all luns have all the capabilities
        associations = []
        for lun in requested_luns:
            for capability in Entities.get_capability_iterator():
                capability_id = Entities.unique_capability_id(capability)
                associations.append((id_builder(lun), capability_id))
        self._create_associations(associations)
