from ZSI import FaultException
from .zsi.vasaService_types import ns2
from .entities import Entities
from .utils import decorators
from logging import getLogger
logger = getLogger(__name__)

class VasaProviderEntityQueryMixin(object):
    """ implements queryArrays, queryStorageLuns, queryStorageProcessors,
    queryStorageCapabilities, queryStoragePorts """

    def _storage_array_from_device(self, device):
        storage_array = self.local.response.new_return()
        storage_array.set_element_uniqueIdentifier(Entities.unique_array_id(device))
        storage_array.set_element_arrayName(device.get_vendor().get_system_name())
        storage_array.set_element_vendorId(device.get_scsi_vendor_id())
        storage_array.set_element_modelId(device.get_scsi_product_id())
        storage_array.set_element_firmware(device.get_vendor().get_system_version())
        storage_array.set_element_supportedBlock(["FC"])
        storage_array.set_element_supportedFileSystem([])
        storage_array.set_element_alternateName([])
        storage_array.set_element_supportedProfile(["BlockDeviceProfile", "CapabilityProfile"])
        return storage_array

    def _volume_from_device(self, device):
        infinisdk_system = Entities.get_system_object_by_device(device)
        if infinisdk_system is None:
            return None
        infinisdk_volume = infinisdk_system.volumes.safe_get(id=device.get_vendor().get_volume_id())
        return infinisdk_volume

    def _storage_lun_from_device(self, device):
        from capacity import MB
        naa = Entities.unique_lun_id(device)
        infinisdk_volume = self._volume_from_device(device)
        if infinisdk_volume is None:
            logger.warn("Could not get volume for device {}".format(naa))
            return None
        storage_lun = self.local.response.new_return()
        storage_lun.set_element_uniqueIdentifier(naa)
        storage_lun.set_element_esxLunIdentifier(naa)
        storage_lun.set_element_displayName(infinisdk_volume.get_name())
        if infinisdk_volume.get_provisioning() == "THIN":
            storage_lun.set_element_thinProvisioned(True)
            storage_lun.set_element_thinProvisioningStatus("Green")
        else:
            storage_lun.set_element_thinProvisioned(False)
            # not calling storage_lun.set_element_thinProvisioningStatus
        storage_lun.set_element_capacityInMB(infinisdk_volume.get_size() / MB)
        storage_lun.set_element_usedSpaceInMB(infinisdk_volume.get_allocated() / MB)
        storage_lun.set_element_alternateIdentifier([infinisdk_volume.get_name()])
        storage_lun.set_element_drsManagementPermitted(True)
        return storage_lun

    def _storage_processor_from_node(self, node):
        storage_processor = self.local.response.new_return()
        storage_processor.set_element_uniqueIdentifier(Entities.unique_node_id(node))
        storage_processor.set_element_spIdentifier(["Node-{}".format(node.get_id())])
        return storage_processor

    def _storage_port_from_port(self, port):
        storage_port = self.local.response.new_return()
        storage_port.set_element_uniqueIdentifier(Entities.unique_port_id(port))
        storage_port.set_element_portType("FC")
        storage_port.set_element_portWwn(port.port_wwn)
        storage_port.set_element_nodeWwn(port.node_wwn)
        # not calling storage_port.set_element_iscsiIdentifier()
        storage_port.set_element_alternateName([])
        return storage_port

    def _storage_capability_from_capability(self, capability):
        capability_name, capability_detail = capability
        storage_capability = self.local.response.new_return()
        storage_capability.set_element_uniqueIdentifier(Entities.unique_capability_id(capability))
        storage_capability.set_element_capabilityName(capability_name)
        storage_capability.set_element_capabilityDetail(capability_detail)
        return storage_capability

    def _request_ids_to_items(self, requested_ids, iterator, id_builder):
        """ Compare all known items (returned from iterator) with an array requested by VASA query functions.
        Returns requested items, according to the uniqueIds in the array """
        id_to_item = {id_builder(item): item for item in iterator}
        check_ever_returned = True
        if all(requested_id is None for requested_id in requested_ids):
            # empty or all None - means return all
            requested_ids = id_to_item.keys()
            check_ever_returned = False
        for requested_id in requested_ids:
            if check_ever_returned and requested_id not in self.ids_ever_returned:
                raise FaultException(ns2.InvalidArgument_Dec().pyclass())
            if requested_id not in id_to_item:
                continue        # according to the spec, stale ids should be ignored
            item = id_to_item[requested_id]
            yield item

    def _query_response_builder(self, requested_ids, iterator, id_builder, item_builder):
        """ base function for queryArrays and queryStorage* """
        # iterator is one of _get_*_iterator (e.g. devices, ports...)
        # id_builder returns the uniqueId for each element, as returned by the provider in
        # queryUniqueIdentifiersForEntity, so the ids we'll get back from the vCenter will be in this format.
        # item_builder returns a VASA data object for an element returned by iterator
        return_array = []
        for item in self._request_ids_to_items(requested_ids, iterator, id_builder):
            self.ids_ever_returned.add(id_builder(item))
            result = item_builder(item)
            if result is None:
                continue
            return_array.append(result)
        self.local.response.set_element_return(return_array)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryArrays(self):
        requested_ids = self.local.request.get_element_arrayUniqueId()
        iterator = Entities.get_controller_iterator()
        id_builder = Entities.unique_array_id
        item_builder = self._storage_array_from_device
        self._query_response_builder(requested_ids, iterator, id_builder, item_builder)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryStorageLuns(self):
        requested_ids = self.local.request.get_element_lunUniqueId()
        if not requested_ids:
            # all other query functions support "empty" as "all", but not this one, so we check for empty parameter here
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())
        iterator = Entities.get_device_iterator()
        id_builder = Entities.unique_lun_id
        item_builder = self._storage_lun_from_device
        self._query_response_builder(requested_ids, iterator, id_builder, item_builder)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryStorageProcessors(self):
        requested_ids = self.local.request.get_element_spUniqueId()
        iterator = Entities.get_node_iterator()
        id_builder = Entities.unique_node_id
        item_builder = self._storage_processor_from_node
        self._query_response_builder(requested_ids, iterator, id_builder, item_builder)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryStorageCapabilities(self):
        iterator = Entities.get_capability_iterator()
        id_builder = Entities.unique_capability_id
        item_builder = self._storage_capability_from_capability
        requested_ids = self.local.request.get_element_capabilityUniqueId()
        self._query_response_builder(requested_ids, iterator, id_builder, item_builder)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryStoragePorts(self):
        iterator = Entities.get_port_iterator()
        id_builder = Entities.unique_port_id
        item_builder = self._storage_port_from_port
        requested_ids = self.local.request.get_element_portUniqueId()
        self._query_response_builder(requested_ids, iterator, id_builder, item_builder)
