from ZSI import FaultException
from .zsi.vasaService_types import ns2
from .utils import decorators, client_certificates, client_sessions, local_storage
from infi.pyutils.contexts import contextmanager
from logging import getLogger
logger = getLogger(__name__)

class VasaProviderRegistrationMixin(object):
    """ implements registerVASACertificate, unregisterVASACertificate, setContext """

    def _fill_provider_info(self, result, session):
        from infi.vendata.vmware_powertools import info
        from infi.storagemodel.vendor.infinidat.infinibox import vid_pid_with_spaces as vid_pid
        result.set_element_name(info.get_product_name())
        result.set_element_defaultNamespace(vid_pid[0])
        result.set_element_defaultSessionTimeoutInSeconds(client_sessions.ActiveSessionsPool.TIMEOUT_IN_SECONDS)
        if session.get_session_id():
            result.set_element_sessionId(str(session.get_session_id()))
        svm = result.new_supportedVendorModel()
        svm.set_element_vendorId(vid_pid[0])
        svm.set_element_modelId(vid_pid[1])
        result.set_element_supportedVendorModel([svm])
        result.set_element_vasaApiVersion("1.0")
        result.set_element_vasaProviderVersion(info.get_product_version())

    def _validate_credentials(self):
        from infi.vendata.vmware_powertools.config import Configuration
        expected_username, expected_password = Configuration.get_vasa_provider_credentials()
        username, password = self.local.request.get_element_userName(), self.local.request.get_element_password()
        if username != expected_username or password != expected_password:
            raise FaultException(ns2.InvalidLogin_Dec().pyclass(),
                                 string="Cannot complete login due to an incorrect user name or password")

    def _add_new_certificate_to_trust_store(self):
        newCertificate = self.local.request.get_element_newCertificate()
        certificate = client_certificates.Certificate.from_vasa_api(newCertificate)
        client_certificates.TrustedStore.add(certificate)

    def _invalidate_old_certificate(self):
        certficate = client_certificates.Certificate.from_request(local_storage.get_request())
        client_certificates.TrustedStore.remove(certficate)

    @contextmanager
    def _certificate_in_file_context(self, certificate_content):
        import tempfile
        import os
        fd, certificate_file = tempfile.mkstemp()
        os.write(fd, client_certificates.Certificate.base64_encoded_der_to_pem_format(certificate_content or ''))
        os.close(fd)
        try:
            yield certificate_file
        finally:
            os.remove(certificate_file)

    def _read_certificate_expiry_date_from_file(self, certificate_file):
        import infi.execute
        try:
            cmd = ['openssl', 'x509', '-in', certificate_file, '-noout', '-enddate']
            pid = infi.execute.execute_assert_success(cmd)  # prints notAfter=Dec 17 08:21:32 2112 GMT
        except infi.execute.ExecutionError:
            with open(certificate_file) as fd:
                newCertificate = fd.read()
            logger.warn("newCertificate {!r} is not a valid certficate".format(newCertificate))
            raise FaultException(ns2.InvalidCertificate_Dec().pyclass(),
                                 string="certificate is not in a valid PEM format")
        try:
            output = pid.get_stdout().strip()
            not_after = output.splitlines()[0].split('=')[-1]
            return client_certificates.Certificate.not_after_string_to_datetime_object(not_after)
        except Exception:
            logger.warn("Failed to parse certificate expiry date from {!r}".format(pid.get_stdout()))
            raise FaultException(ns2.InvalidCertificate_Dec().pyclass(),
                             string="could not parse notAfter attribute from certificate")

    def _validate_certificate(self, certficate):
        import datetime
        with self._certificate_in_file_context(certficate) as certificate_file:
            expiry_date = self._read_certificate_expiry_date_from_file(certificate_file)
            if expiry_date < datetime.datetime.utcnow():
                logger.warn("certficate is expired")
                raise FaultException(ns2.InvalidCertificate_Dec().pyclass(),
                                     string="certificate is expired")

    def _is_certificate_valid(self, certficate):
        try:
            self._validate_certificate(certficate)
            return True
        except FaultException:
            return False

    @decorators.soap_decorator
    def soap_registerVASACertificate(self):
        self._validate_credentials()
        newCertificate = self.local.request.get_element_newCertificate()
        self._validate_certificate(newCertificate)
        self._add_new_certificate_to_trust_store()
        result = self.local.response.new_return()
        self._fill_provider_info(result, client_sessions.Session.from_handler(local_storage.get_handler()))
        self.local.response.set_element_return(result)

    @decorators.soap_decorator
    def soap_unregisterVASACertificate(self):
        existingCertificate = self.local.request.get_element_existingCertificate()
        self._validate_certificate(existingCertificate)
        certificate = client_certificates.Certificate.from_vasa_api(existingCertificate)
        if not client_certificates.TrustedStore.is_trusted(certificate):
            if client_certificates.TrustedStore.is_revoked(certificate):
                raise FaultException(ns2.InvalidSession_Dec().pyclass(),
                                     string="Invalid session, certificate is revoked")
            raise FaultException(ns2.InvalidCertificate_Dec().pyclass(),
                                 string="certificate not registered")
        session = client_sessions.get_session()
        self._deinit_alarm_session(session.get_session_id())
        self._deinit_event_session(session.get_session_id())
        client_sessions.ActiveSessionsPool.remove(session)
        client_certificates.TrustedStore.remove(certificate)

    @decorators.validate_certificate
    @decorators.soap_decorator
    def soap_setContext(self):
        usageContext = self.local.request.get_element_usageContext()
        if usageContext is None:
            raise FaultException(ns2.InvalidArgument_Dec().pyclass(),
                                 string="usageContext is None")
        session = client_sessions.Session.generate_new()
        session_timeout = usageContext.get_element_sessionTimeoutInSeconds()
        if session_timeout:
            session.set_session_timeout(session_timeout)
        host_initiator = usageContext.get_element_hostInitiator()
        session.set_session_value('host_initiator', host_initiator)
        client_sessions.ActiveSessionsPool.add(session)
        session_id = session.get_session_id()
        local_storage.get_handler().add_additional_header('Set-Cookie', "VASASESSIONID={}".format(session_id))
        local_storage.set_session(session)
        self._init_alarm_session(session_id)
        self._init_event_session(session_id)
        result = self.local.response.new_return()
        self._fill_provider_info(result, session)
        self.local.response.set_element_return(result)
