from infi.pyutils.decorators import wraps
from . import local_storage, client_certificates, client_sessions
from ZSI import FaultException
from ..zsi import vasaService_client
from ..zsi.vasaService_types import ns2
from logging import getLogger
logger = getLogger(__name__)


def validate_certificate(func):
    def _get_certificate():
        request = local_storage.get_request()
        certificate = client_certificates.Certificate.from_request(request)
        return certificate

    def _raise_if_not_trusted(certificate):
        if not client_certificates.TrustedStore.is_trusted(certificate):
            logger.error("certificate {0} is not trusted".format(certificate.get_id()))
            raise FaultException(ns2.InvalidSession_Dec().pyclass(),
                                 string="certificate is not trusted")

    def _raise_if_expired(certificate):
        if certificate.is_expired():
            client_certificates.TrustedStore.remove(certificate)
            logger.warn("certificate {0} is expired".format(certificate.get_id()))
            raise FaultException(ns2.InvalidSession_Dec().pyclass(),
                                 string="certificate is expired")

    @wraps(func)
    def decorator(*args, **kwargs):
        certificate = _get_certificate()
        _raise_if_not_trusted(certificate)
        _raise_if_expired(certificate)
        return func(*args, **kwargs)
    return decorator


def validate_session(func):
    def _raise_if_invalid(session):
        if not client_sessions.ActiveSessionsPool.is_trusted(session):
            logger.error("session {0} is not trusted".format(session.get_id()))
            raise FaultException(ns2.InvalidSession_Dec().pyclass(),
                                 string="session is not trusted")

    def _raise_if_expired(session):
        if client_sessions.ActiveSessionsPool.is_expired(session):
            client_sessions.ActiveSessionsPool.remove(session)
            logger.warn("session {0} is expired".format(session.get_id()))
            raise FaultException(ns2.InvalidSession_Dec().pyclass(),
                                 string="session is expired")

    @wraps(func)
    def decorator(*args, **kwargs):
        session = client_sessions.get_session()
        _raise_if_invalid(session)
        _raise_if_expired(session)
        client_sessions.ActiveSessionsPool.update(session)
        return func(*args, **kwargs)
    return decorator


def soap_decorator(meth):
    """ assumes instance has "local" variable. fills it with the correct request and response objects
    of the correct SOAP type """
    @wraps(meth)
    def new_meth(inst, ps):
        action = meth.func_name[5:]  # strip "soap_"
        request_class = getattr(vasaService_client, action + "Request", None)
        response_class = getattr(vasaService_client, action + "Response", None)
        inst.local.request = ps.Parse(request_class.typecode) if request_class is not None else None
        inst.local.response = response_class() if response_class is not None else None
        meth(inst)
        return inst.local.request, inst.local.response
    return new_meth


def soap_not_implemented(meth):
    @wraps(meth)
    def new_meth(*args, **kwargs):
        raise FaultException(ns2.NotImplemented_Dec().pyclass())
    return new_meth
