import time
import os
from ZSI import FaultException
from .zsi.vasaService_types import ns2
from .zsi.vasaService_server import vasaService
from .utils import decorators
from .entity_associations import VasaProviderEntityAssociationMixin
from .entity_queries import VasaProviderEntityQueryMixin
from .events import VasaProviderEventsMixin
from .alarms import VasaProviderAlarmsMixin
from .registration import VasaProviderRegistrationMixin
from .entities import Entities, ARRAY_ID_PREFIX
from logging import getLogger
logger = getLogger(__name__)

class Service(VasaProviderEventsMixin, VasaProviderRegistrationMixin,
              VasaProviderEntityQueryMixin, VasaProviderEntityAssociationMixin, VasaProviderAlarmsMixin,
              vasaService):

    # map entity type -> 2-tuple of (entity iterator, entity id builder)
    ENTITY_MAP = {"StorageProcessor": (Entities.get_node_iterator, Entities.unique_node_id),
                  "StoragePort": (Entities.get_port_iterator, Entities.unique_port_id),
                  "StorageCapability": (Entities.get_capability_iterator, Entities.unique_capability_id),
                  "StorageLun": (Entities.get_device_iterator, Entities.unique_lun_id),
                  "StorageArray": (Entities.get_controller_iterator, Entities.unique_array_id)}

    def __init__(self, *args, **kwargs):
        from threading import local
        self.ids_ever_returned = set()
        self.local = local()
        VasaProviderEventsMixin.__init__(self)
        VasaProviderAlarmsMixin.__init__(self)
        vasaService.__init__(self, *args, **kwargs)               # old-style class

    def _create_catalog(self, name, uri, modified_time):
        from infi.storagemodel.vendor.infinidat.infinibox import vid_pid
        result = self.local.response.new_return()
        result.set_element_catalogName(name)
        result.set_element_catalogUri(uri)
        result.set_element_catalogVersion("1.0")
        result.set_element_lastModified(modified_time)
        result.set_element_locale("en")
        result.set_element_moduleName(vid_pid[0])
        return result

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryCatalog(self):
        file_modified_time = time.gmtime(os.stat(__file__).st_mtime)
        result1 = self._create_catalog("Alarm", "/alarm.vmsg", file_modified_time)
        result2 = self._create_catalog("Event", "/event.vmsg", file_modified_time)
        result3 = self._create_catalog("Fault", "/fault.vmsg", file_modified_time)
        self.local.response.set_element_return([result1, result2, result3])

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryDRSMigrationCapabilityForPerformance(self):
        src_id = self.local.request.get_element_srcUniqueId()
        dst_id = self.local.request.get_element_dstUniqueId()
        entity_type = self.local.request.get_element_entityType()
        if src_id is None or dst_id is None or entity_type is None:
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())
        if not src_id.startswith("naa.") or not dst_id.startswith("naa.") or entity_type != "StorageLun":
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())
        if src_id not in self.ids_ever_returned or dst_id not in self.ids_ever_returned:
            raise FaultException(ns2.NotFound_Dec().pyclass())
        self.local.response.set_element_return(True)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_getNumberOfEntities(self):
        def ilen(iter):
            """ return iterator length """
            return sum(1 for _ in iter)
        entity_type = self.local.request.get_element_entityType()
        if entity_type not in self.ENTITY_MAP.keys():
            # StorageFileSystem or unknown
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())
        entity_iterator, entitiy_id_builder = self.ENTITY_MAP[entity_type]
        self.local.response.set_element_return(ilen(entity_iterator()))

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryUniqueIdentifiersForEntity(self):
        entity_type = self.local.request.get_element_entityType()
        if entity_type not in self.ENTITY_MAP.keys():
            # StorageFileSystem or unknown
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())
        entity_iterator, entitiy_id_builder = self.ENTITY_MAP[entity_type]
        res = [entitiy_id_builder(entity) for entity in entity_iterator()]
        self.ids_ever_returned.update(set(res))
        self.local.response.set_element_return(res)

    @decorators.validate_certificate
    @decorators.validate_session
    @decorators.soap_decorator
    def soap_queryUniqueIdentifiersForLuns(self):
        system_serial = self.local.request.get_element_arrayUniqueId()
        if not system_serial or                                 \
            not system_serial.startswith(ARRAY_ID_PREFIX) or    \
            system_serial not in self.ids_ever_returned:
            raise FaultException(ns2.InvalidArgument_Dec().pyclass())
        system_serial = int(system_serial[len(ARRAY_ID_PREFIX):])
        device_iterator = Entities.get_device_iterator()
        res = [Entities.unique_lun_id(dev) for dev in device_iterator if
               dev.get_vendor().get_system_serial() == system_serial]
        self.ids_ever_returned.update(set(res))
        self.local.response.set_element_return(res)

    @decorators.soap_not_implemented
    def soap_queryAssociatedCapabilityForFileSystem(self):
        pass

    @decorators.soap_not_implemented
    def soap_queryUniqueIdentifiersForFileSystems(self):
        pass

    @decorators.soap_not_implemented
    def soap_queryStorageFileSystems(self):
        pass
