from infi.pyutils.lazy import cached_function
from pyVmomi import vim

DEFAULT_POST_URL = "https://{}/vsm/extensionService"
REGISTER_POST_DATA = """<RegisterExtension xmlns="http://www.vmware.com/schema/vservice/ExtensionVService">
       <Key>{key}</Key>
       <Certificate>
{certificate}
       </Certificate>
</RegisterExtension>
"""


from logging import getLogger
logger = getLogger(__name__)

VCENTER_CLIENT_KEEPALIVE_IN_SECONDS = 60 * 5


class vCenter(object):
    _impersonated_clients = {}

    @classmethod
    def clear_cache_for_get_client_methods(cls):
        from infi.pyutils.lazy import clear_cache
        clear_cache(cls.get_extension_client)
        clear_cache(cls.get_impersonated_client_by_session_id)

    @classmethod
    def get_vcenter_client(cls):
        from infi.pyvmomi_wrapper import Client
        from ..config import Configuration
        vcenter_address = Configuration.get_vcenter_address()
        client = Client(vcenter_address,
                        **Configuration.get_vcenter_extension_ssl_configuration())
        client.login_extension_by_certificate(Configuration.get_vcenter_extension_key())
        logger.info("Client created")
        return client

    @classmethod
    def _is_there_an_update(cls, client):
        return any(value.check_for_updates() for value in client.property_collectors.values())

    @classmethod
    def _start_keepalive_thread_for_client(cls, client):
        from gevent import sleep, GreenletExit
        from ..scripts.starter import _populate_storagemodel
        logger.info("vcenter client keepalive thread started")
        while True:
            sleep(VCENTER_CLIENT_KEEPALIVE_IN_SECONDS)
            try:
                if cls._is_there_an_update(client):
                    _populate_storagemodel()
            except:
                logger.exception("Exception has occurred in the vcenter client keepalive thread")
                try:
                    cls.reconnect()
                except GreenletExit:
                    raise
                except:
                    logger.exception("Failed to reconnect from the keepalive thread")

    @classmethod
    def stop_keepalive_greenlet(cls):
        try:
            if hasattr(cls, "_keepalive_greenlet"):
                cls._keepalive_greenlet.kill()
        except:
            logger.exception("Killing the keepalive greenlet raised an exception")

    @classmethod
    @cached_function
    def get_extension_client(cls):
        from infi.vendata.vmware_powertools.utils import spawn
        logger.info("Getting a pyVmomi client for the first time")
        client = cls.get_vcenter_client()
        cls._keepalive_greenlet = spawn(cls._start_keepalive_thread_for_client, client)
        return client

    @classmethod
    def prepare_property_collectors(cls):
        from ..patches import install_all_property_collectors_on_client, fetch_all_property_collectors
        client = cls.get_extension_client()
        install_all_property_collectors_on_client(client)
        fetch_all_property_collectors(client)

    @classmethod
    def reconnect(cls):
        from infi.vendata.vmware_powertools.utils import spawn
        spawn(cls._start_reconnect_thread).join()

    @classmethod
    def _start_reconnect_thread(cls):
        from infi.storagemodel.vmware.patches.storagemodel import StorageModelFactory
        logger.info("Re-connecting to vCenter")
        cls.stop_keepalive_greenlet()
        cls.clear_cache_for_get_client_methods()
        StorageModelFactory.clear()
        cls.get_extension_client()
        vCenter.prepare_property_collectors()

    @classmethod
    def get_active_sessions(cls, try_again=True):
        """:returns: a dictionary of active sessions from vcenter
        keys are the session ids, and the value holds the matching usernames"""
        import socket
        from httplib import BadStatusLine
        try:
            return {session.key: session.userName
                    for session in cls.get_extension_client().session_manager.sessionList}
        except (vim.SecurityError, socket.error, BadStatusLine):
            if not try_again:
                raise
            cls.reconnect()
            return cls.get_active_sessions(try_again=False)

    @classmethod
    def get_impersonated_client_by_username(cls, username, locale):
        client = cls.get_vcenter_client()
        client.session_manager.ImpersonateUser(username, locale)
        return client

    @classmethod
    @cached_function
    def get_impersonated_client_by_session_id(cls, session_id, locale):
        # TODO we need to clean the cache for this function every once in a while
        username = cls.get_active_sessions()[session_id]
        return cls.get_impersonated_client_by_username(username, locale)

    @classmethod
    def get_vcenter_attributes(cls):
        from ..ovf import OVF
        return OVF.get_attributes()

    @classmethod
    def register_extension_through_vservice(cls):
        from infi.vendata.vmware_powertools.scripts.greeter import ConnectivityError
        cls._send_post_request_to_vcenter()
        logger.info("Registered vCenter extension")
        try:
            cls._update_extension()
        except Exception, error:
            logger.exception("failed to update vCenter extension")
            raise ConnectivityError(error)
        # TODO UpdateExtension deletes the certificate and we cannot login
        cls._send_post_request_to_vcenter()
        logger.info("Extension updated")

    @classmethod
    def update_vm_version_if_necessary(cls):
        from infi.vendata.vmware_powertools.info.update_vm import update_vm_version_if_necessary
        update_vm_version_if_necessary()

    @classmethod
    def register_vasa_provider(cls):
        from ..config import Configuration
        from infi.pyvmomi_wrapper.sms import SmsClient
        from pyVmomi import sms
        client = cls.get_extension_client()
        sms_client = SmsClient(client)
        name = Configuration.get_vcenter_extension_key()
        try:
            storage_manager = sms_client.service_instance.QueryStorageManager()
        except:
            logger.exception("VASA APIs do not exist")
            return

        providers = storage_manager.QueryProvider()
        for provider_info in [provider.QueryProviderInfo() for provider in providers]:
            if provider_info.name == name:
                storage_manager.UnregisterProvider_Task(provider_info.uid)
        username, password = Configuration.get_vasa_provider_credentials()
        spec = sms.provider.VasaProviderSpec(url=Configuration.get_vasa_provider_url(),
                                             name=name, username=username, password=password,
                                             certificate=cls._get_webserver_certificate())
        logger.info("vasa spec: {!r}".format(spec))
        try:
            task = storage_manager.RegisterProvider_Task(spec)
            sms_client.wait_for_task(task)
        except:
            logger.exception("failed to register vasa prvider")
        else:
            logger.info("VASA provider registered")

    @classmethod
    def _prepare_headers(cls):
        attributes = cls.get_vcenter_attributes()
        return {'evs-token': attributes['Token']}

    @classmethod
    def _prepare_data(cls):
        from ..config import Configuration
        ssl = Configuration.get_vcenter_extension_ssl_configuration()
        with open(ssl['certfile']) as fd:
            certificate = fd.read()
        return REGISTER_POST_DATA.format(certificate=certificate,
                                         key=Configuration.get_vcenter_extension_key())

    @classmethod
    def _send_post_request_to_vcenter_by_address(cls, address):
        from urllib2 import urlopen, Request
        request = Request(address,
                          headers=cls._prepare_headers(),
                          data=cls._prepare_data())
        return urlopen(request)

    @classmethod
    def _send_post_request_to_vcenter(cls):
        # vsphere-ext-solutions-50.pdf, page 82
        # Developing and Deploying vSphere Solutions, vServices, and ESX Agents
        # Section 6 Integrating an Extension with the vCenter Extension vService
        # Provide a Script in the Exetension Virtual Machine to Register as a vCenter Extension
        from infi.vendata.vmware_powertools.scripts.greeter import vServiceRegistrationFailure
        vcenter_attributes = cls.get_vcenter_attributes()
        first_url = '<missing-in-ovf>'
        second_url = '<missing-in-ovf>'
        try:
            if 'URL' not in vcenter_attributes:
                raise Exception("vCenter URL not present in OVF")
            first_url = cls.get_vcenter_attributes()['URL']
            if not first_url:
                raise Exception("vCenter URL is empty in OVF")
            return cls._send_post_request_to_vcenter_by_address(first_url)
        except Exception, first_http_error:
            logger.exception("vService registration by URL failed, failing back to IP-based registration")
        try:
            if 'IP' not in vcenter_attributes:
                raise Exception("vCenter IP not present in OVF")
            if not vcenter_attributes['IP']:
                raise Exception("vCenter IP is empty in OVF")
            second_url = DEFAULT_POST_URL.format(vcenter_attributes['IP'])
            cls._send_post_request_to_vcenter_by_address(second_url)
        except Exception, second_http_error:
            logger.exception("vService registration by IP, aborting")
            raise vServiceRegistrationFailure((first_url, first_http_error), (second_url, second_http_error))

    @classmethod
    def _get_webserver_certificate(cls):  # pragma: no cover
        from ..config import Configuration
        with open(Configuration.get_webserver_ssl_configuration()['certfile']) as fd:
            content = fd.read()
        return ''.join(content.splitlines()[1:-1])

    @classmethod
    def _get_webserver_ssl_certficate_thumbprint(cls):  # pragma: no cover
        from infi.execute import execute_assert_success
        from ..config import Configuration
        openssl = execute_assert_success(['openssl', 'x509', '-fingerprint', '-noout', '-in',
                                          Configuration.get_webserver_ssl_configuration()['certfile']])
        # SHA1 Fingerprint=57:EF:96:5F:9E:EF:7A:59:19:BA:13:2C:1B:4D:72:D7:00:03:3D:EC
        return openssl.get_stdout().split('=')[1].strip()

    @classmethod
    def _update_extension(cls):
        from .. import info, config
        URL = '/'.join([config.Configuration.get_webapp_url(), 'plugin/info'])
        client = cls.get_extension_client()
        description = vim.Description(label=info.get_product_description(),
                                      summary=info.get_product_name())
        extension_manager = client.service_content.extensionManager
        extension = extension_manager.FindExtension(config.Configuration.get_vcenter_extension_key())
        extension.client = []
        kwargs = dict(adminEmail=info.get_company_email(), company=info.get_company_name(),
                      description=description, type=info.get_plugin_type(), url=URL,
                      serverThumbprint=cls._get_webserver_ssl_certficate_thumbprint())
        extension.server = [vim.ExtensionServerInfo(**kwargs)]
        extension.company = info.get_company_name()
        extension.description = description
        extension.subjectName = config.Configuration.get_certificate_subject_name()
        extension.type = info.get_plugin_type()
        extension.version = info.get_product_version()
        extension.extendedProductInfo = None
        extension.solutionManagerInfo = None
        client.service_content.extensionManager.UpdateExtension(extension)
