from logging import getLogger

logger = getLogger(__name__)

class NoSystemFoundException(Exception):
    pass

class HostSystemByController(object):
    def __init__(self, esx_host, system_serial):
        super(HostSystemByController, self).__init__()
        self._esx_host = esx_host
        self._system_serial = system_serial

    def get_controller_device(self):
        from infi.vendata.vmware_powertools.patches import with_host
        with with_host(self._esx_host):
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
            for device in get_infinidat_block_devices_and_controllers__mapped_to_lun0():
                if device.get_vendor().get_system_serial() == self._system_serial:
                    return device
        template = "Couldn't find storage controller for system {} on host {}"
        message = template.format(self._system_serial, self._esx_host.summary.config.name)
        raise NoSystemFoundException(message)

    def get_system(self):
        from infi.vendata.vmware_powertools.patches import with_host
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device
        device = self.get_controller_device()
        with with_host(self._esx_host):
            return get_system_object_by_device(device)

    def get_host(self):
        device = self.get_controller_device()
        system = self.get_system()
        host = system.hosts.get(id=device.get_vendor().get_host_id())
        return host

    def get_cluster(self):
        device = self.get_controller_device()
        system = self.get_system()
        cluster = system.host_clusters.get(id=device.get_vendor().get_cluster_id())
        return cluster

class MapVolume(HostSystemByController):
    def __init__(self, esx_host, system_serial, volume_id):
        super(MapVolume, self).__init__(esx_host, system_serial)
        self._volume_id = volume_id

    def get_volume(self):
        system = self.get_system()
        volume = system.volumes.get(id=self._volume_id)
        return volume

    def _is_mounted(self, datastore, esx_host):
        return any(host_mount.key._moId == esx_host._moId and host_mount.mountInfo.mounted
                   for host_mount in datastore.host)

    def mount_datastore(self):
        # we want to reverse unmap, which calls both UnmountVmfsVolume and DetachScsiLun
        # AttachScsiLun is automatically called by the rescan (which is called in map_existing_volume)
        # we still need to mount the volumes. we go over datastores under the newly mapped device and mount them
        from infi.vendata.vmware_powertools.vcenter.shortcuts import iter_infinidat_datastores
        from infi.vendata.vmware_powertools.patches import with_host
        with with_host(self._esx_host):
            for datastore, infinidat_devices in iter_infinidat_datastores(self._esx_host):
                if self._volume_id in [infinidat_device.get_vendor().get_volume_id() for infinidat_device
                                       in infinidat_devices]:
                    if not self._is_mounted(datastore, self._esx_host):
                        self._esx_host.configManager.storageSystem.MountVmfsVolume(datastore.info.vmfs.uuid)

    def map_to_host(self):
        host, volume = self.get_host(), self.get_volume()
        logger.info("Mapping volume {} to host {}".format(volume.get_name(), host.get_name()))
        host.map_volume(volume=volume)

    def map_to_cluster(self):
        cluster, volume = self.get_cluster(), self.get_volume()
        logger.info("Mapping volume {} to cluster {}".format(volume.get_name(), cluster.get_name()))
        cluster.map_volume(volume=volume) if volume not in cluster.get_luns() else None
