from logging import getLogger
from gevent.lock import Semaphore as Lock
logger = getLogger(__name__)

taks_lock = Lock()

def create_vcenter_task(parent_object, task_name):
    from infi.vendata.vmware_powertools.config import Configuration
    from infi.pyvmomi_wrapper import TaskManager
    from infi.vendata.vmware_powertools.vcenter import vCenter
    client = vCenter.get_extension_client()
    key = Configuration.get_vcenter_extension_key()
    # we use a lock to prevent race where two gevents register the same task type and run over each other
    taks_lock.acquire()
    try:
        task = TaskManager(client, key).task(parent_object, task_name)
    finally:
        taks_lock.release()
    return task

def moref_to_object(moref):
    from infi.vendata.vmware_powertools.vcenter import vCenter
    client = vCenter.get_extension_client()
    vcenter_object = client.get_managed_object_by_reference(moref)
    return vcenter_object

def object_to_moref(mo):
    from infi.pyvmomi_wrapper import get_reference_to_managed_object
    return get_reference_to_managed_object(mo)

def wait_for_task(task, timeout=None):
    from infi.vendata.vmware_powertools.vcenter import vCenter
    client = vCenter.get_extension_client()
    client.wait_for_task(task, timeout)

def get_infinidat_devices_under_datastore(datastore):
    from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
    underlying_disk_names = [extent.diskName for extent in datastore.info.vmfs.extent]
    return [device for device in get_infinidat_block_devices()
            if device.get_canonical_name() in underlying_disk_names]

def get_datastores_on_block_storage(esx_host):
    return [datastore for datastore in esx_host.datastore if datastore.summary.type == 'VMFS']

def iter_infinidat_datastores(esx_host):
    for datastore in get_datastores_on_block_storage(esx_host):
        infinidat_devices = get_infinidat_devices_under_datastore(datastore)
        if infinidat_devices:
            yield datastore, infinidat_devices

def get_host_by_uuid(client, uuid):
    return client.service_content.searchIndex.FindByUuid(uuid=uuid, vmSearch=False)

def get_vm_by_uuid(client, uuid):
    return client.service_content.searchIndex.FindByUuid(uuid=uuid, vmSearch=True)

def get_datastore_by_url(client, url):
    matches = [ds for ds in client.get_datastores() if ds.info.url == url]
    if len(matches) == 0:
        return None
    return matches[0]

def get_datacenter_clusters(datacenter):
    from pyVmomi import vim
    return [mo for mo in datacenter.hostFolder.childEntity if isinstance(mo, vim.ClusterComputeResource)]

def get_cluster_datastores(cluster):
    # ComputeResource.datastore documentation: "This property is computed as the aggregate set of datastores available
    # from all the hosts that are part of this compute resource."
    # i.e. cluster.datastore also returns datastores attached to only a subset of the hosts
    # compute the datastores that all hosts can see (intersecting set), without relying on cluster.datastore
    if not cluster.host:
        return set()
    owner_host_datastores = [set(host.datastore) for host in cluster.host]
    return set.intersection(*owner_host_datastores)


# Tree under serviceContent
# -> (Folder) rootFolder
#  |-> (Datacenter[]) childEntity
#    |-> (Datastore[]) datastore
#    |-> (Folder) hostFolder
#      |-> (ClusterComputeResource[]) childEntity
#        |-> (Datastore[]) datastore
#          |-> (VmfsDatastoreInfo) info
#            |-> (HostVmfsVolume) vmfs
#              |-> (HostScsiDiskPartition[]) extent
#                |-> (string) diskName
#        |-> (HostSystem[]) host
#          |-> (Datastore[]) datastore
#          |-> (HostConfigInfo) config
#            |-> (HostStorageDeviceInfo) storageDevice
#              |-> (HostHostBusAdapter[]) hostBusAdater
#              |-> (HostMultipathInfo) multipathInfo
#                |-> (HostMultipathInfoLogicalUnit[]) lun
#                  |-> (HostMultipathInfoPath[]) path
#              |-> (HostPlugStoreTopology) plugStoreTopology
#              |-> (ScsiLun[]) scsiLun
#                |-> (string) canonicalName
#                |-> (ScsiLunDurableName[]) alternateName
#              |-> (HostScsiTopology) scsiTopology

# HostScsiDiskPartition.diskName == ScsiLun.canonicalName
# ScsiLun.alternateName holds VPDs
# HostMultipathInfoLogicalUnit.lun == HostScsiDisk.key
