from infi.vendata.vmware_powertools.patches import with_host
from infi.vendata.vmware_powertools.vcenter.shortcuts import wait_for_task, object_to_moref, moref_to_object
from infi.pyvmomi_wrapper.property_collector import VirtualMachinePropertyCollector
from logging import getLogger
from infi.pyutils.lazy import cached_method
from pyVmomi import vim


logger = getLogger(__name__)


HARDWARE_DEVICES_PATH = "config.hardware.device"


class ValidationError(Exception):
    pass


class UnmapError(Exception):
    pass


class DatastoreAssertions(object):
    def __init__(self, esx_host, datastore):
        super(DatastoreAssertions, self).__init__()
        self._esx_host = esx_host
        self._datastore = datastore

    def _assert_no_virtual_machines_and_templates_associated(self):
        # This checks for all kinds of references (VM located on datastore, VM has DVD image from this datastore)
        if isinstance(self._datastore.vm, list) and len(self._datastore.vm):
            raise ValidationError("VMs are still associated with this datastore")

    def _assert_not_part_of_a_datastore_cluster(self):
        if isinstance(self._datastore.parent, vim.StoragePod):
            raise ValidationError("Datastore is part of Storage Cluster")

    def _assert_not_managed_by_storage_drs(self):
        # Storage DRS is built on top of storage clusters, and we forbid already
        # So there is not point of testing this
        pass

    def _assert_storage_io_control_disabled(self):
        if self._datastore.capability.storageIORMSupported:
            if self._datastore.iormConfiguration and self._datastore.iormConfiguration.enabled:
                raise ValidationError("Storage IO Control is enabled")

    def _assert_distributed_virtual_switch_configuration_file_not_present(self):
        spec = vim.HostDatastoreBrowserSearchSpec(details=None, matchPattern=".dvsData", query=None,
                                                  searchCaseInsensitive=False, sortFoldersFirst=False)
        task = self._datastore.browser.SearchDatastore_Task(datastorePath='[{}]'.format(self._datastore.name),
                                                            searchSpec=spec)
        wait_for_task(task, 10)
        # HostDatastoreBrowserSearchResults
        search_results = task.info.result
        if len(search_results.file):
            raise ValidationError("DVS data file exists on datastore")

    def _get_mount_info(self):
        """:rtype: `HostMountInfo`"""
        for mount in self._datastore.host:
            if object_to_moref(mount.key) == object_to_moref(self._esx_host):
                return mount.mountInfo

    def _is_datastore_mounted(self):
        mount_info = self._get_mount_info()
        return mount_info is not None and mount_info.mounted

    def assert_safe_to_detach(self):
        if not self._is_datastore_mounted():
            return True
        self._assert_no_virtual_machines_and_templates_associated()
        self._assert_not_part_of_a_datastore_cluster()
        self._assert_not_managed_by_storage_drs()
        self._assert_storage_io_control_disabled()
        self._assert_distributed_virtual_switch_configuration_file_not_present()


class VolumeUnmapBase(object):
    # http://kb.vmware.com/selfservice/microsites/search.do?cmd=displayKC&externalId=2004605
    # http://www.lucd.info/2012/04/15/test-if-the-datastore-can-be-unmounted/

    def __init__(self, esx_host, serial, clustered=False):
        super(VolumeUnmapBase, self).__init__()
        self._esx_host = esx_host
        self._serial = serial
        self._clustered = clustered

    @cached_method
    def _get_device(self):
        with with_host(self._esx_host):
            from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_native_multipath_block_devices
            device = [device for device in get_infinidat_native_multipath_block_devices()
                      if device.get_scsi_serial_number() == self._serial]
            if device == []:
                template = "Device {} does not exist on host"
                message = template.format(self._serial, self._esx_host.summary.config.name)
                logger.info(message)
                return None
            device = device[0]
            _ = [device.get_vendor().get_management_address(),   # noqa (caching)
                 device.get_vendor().get_volume_id(), device.get_uuid()]
            return device

    def _get_system(self):
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device
        return get_system_object_by_device(self._device)

    def _get_host_and_volume(self):
        self._device = self._get_device()
        self._system = self._get_system()
        self._host = self._system.hosts.get(id=self._device.get_vendor().get_host_id())
        self._volume = self._system.volumes.get(id=self._device.get_vendor().get_volume_id())
        return self._host, self._volume

    def _get_cluster_and_volume(self):
        self._device = self._get_device()
        self._system = self._get_system()
        self._cluster = self._system.host_clusters.get(id=self._device.get_vendor().get_cluster_id())
        self._volume = self._system.volumes.get(id=self._device.get_vendor().get_volume_id())
        return self._cluster, self._volume

    def _get_datastores(self):
        device = self._get_device()
        datastores = []
        from infi.vendata.vmware_powertools.vcenter.shortcuts import iter_infinidat_datastores
        with with_host(self._esx_host):
            for datastore, infinidat_devices in iter_infinidat_datastores(self._esx_host):
                if device.get_scsi_serial_number() in [x.get_scsi_serial_number() for x in infinidat_devices]:
                    datastores.append(datastore)
            return datastores

    def assert_not_used_by_a_datastore(self):
        for datastore in self._get_datastores():
            DatastoreAssertions(self._esx_host, datastore).assert_safe_to_detach()

    def _is_rdm(self, device):
        return isinstance(device, vim.VirtualDisk) and \
               isinstance(device.backing, vim.VirtualDiskRawDiskMappingVer1BackingInfo) and \
               device.backing.lunUuid == self._get_device().get_uuid()

    def assert_not_being_used_as_rdm(self):
        from infi.vendata.vmware_powertools.vcenter import vCenter
        client = vCenter.get_extension_client()
        collector = VirtualMachinePropertyCollector(client, [HARDWARE_DEVICES_PATH])
        for vm_ref, value_dict in collector.get_properties().items():
            for device in [device for device in value_dict[HARDWARE_DEVICES_PATH] if self._is_rdm(device)]:
                name = moref_to_object(vm_ref).config.name
                msg = "Device is being used as RDM in VirtualMachine {}".format(name)
                raise ValidationError(msg)

    def assert_does_not_device_contain_diagnostic_partition(self):
        for partition in self._esx_host.configManager.diagnosticSystem.QueryAvailablePartition():
            if partition.id.diskName == self._get_device().get_canonical_name():
                msg = "Disk contains diagnostic partition for host {}".format(self._esx_host.summary.config.name)
                raise ValidationError(msg)

    def unmount_datastore(self):
        for datastore in self._get_datastores():
            logger.info("Unmounting datastore {} from host {}".format(datastore, self._esx_host.name))
            datastore_assertions = DatastoreAssertions(self._esx_host, datastore)
            if datastore_assertions._is_datastore_mounted():
                self._esx_host.configManager.storageSystem.UnmountVmfsVolume(datastore.info.vmfs.uuid)

    def detach(self):
        device = self._get_device()
        if device is None:
            raise UnmapError("Could not find device '{}'".format(self._serial))
        logger.info("Detaching LUN {} from host {}".format(device.get_uuid(), self._esx_host.name))
        try:
            self._esx_host.configManager.storageSystem.DetachScsiLun(device.get_uuid())
        except Exception:
            logger.exception("DetachScsiLun failed, we will still proceed with unmapping the volume")

    def unmap(self):
        device = self._get_device()
        if device is None:
            raise UnmapError("Could not find device '{}'".format(self._serial))
        with with_host(self._esx_host):
            if self._clustered:
                # Unmap from cluster
                cluster, volume = self._get_cluster_and_volume()
                logger.info("Unmapping volume {}".format(volume.get_name()))
                for lun_mapping in cluster.get_luns():
                    if lun_mapping.volume == volume and \
                       lun_mapping.clustered:
                        cluster.unmap_volume(volume=volume)
                        break
                else:
                    raise UnmapError("Could not find device to unmap in cluster luns")
            else:
                # Unmap from host
                host, volume = self._get_host_and_volume()
                logger.info("Unmapping volume {}".format(volume.get_name()))
                for lun_mapping in host.get_luns():
                    if lun_mapping.volume == volume and not lun_mapping.clustered:
                        host.unmap_volume(volume=volume)
                        break
                else:
                    raise UnmapError("Could not find device to unmap in host luns")

    def delete(self):
        self._volume.delete()

    def assert_safe_to_remove(self):
        device = self._get_device()
        if device is None:
            return
        self.assert_not_used_by_a_datastore()
        self.assert_not_being_used_as_rdm()
        self.assert_does_not_device_contain_diagnostic_partition()
