import cherrypy
import os
import json
from infi.pyutils.decorators import wraps
from infi.vendata.powertools.gui import webapp as powertools_webapp
from infi.vendata.vmware_powertools import vwc
from jinja2 import Environment, FileSystemLoader

POWERTOOLS_STATIC_PATH = os.path.join(os.path.dirname(powertools_webapp.__file__), 'static')
STATIC_PATH = os.path.join(os.path.dirname(__file__), 'static')
FAVICON_PATH = os.path.join(POWERTOOLS_STATIC_PATH, 'images', 'icon-32.png')

POWERTOOLS_TEMPLATES_PATH = os.path.join(os.path.dirname(powertools_webapp.__file__), 'templates')
TEMPLATES_PATH = os.path.join(os.path.dirname(__file__), 'templates')
TEMPLATE_PATHS = [TEMPLATES_PATH, POWERTOOLS_TEMPLATES_PATH]

VWC_STATIC_PATH = os.path.join(os.path.dirname(vwc.__file__), "plugin", "vwc.zip")

# Environment for loading Jinja templates
JINJA_ENV = Environment(loader=FileSystemLoader(TEMPLATE_PATHS), autoescape=True)

from logging import getLogger
logger = getLogger(__name__)


def is_unexpected_error(error):
    # see documentation in powertools' is_unexpected_error
    from infi.vendata.powertools.gui.webapp import is_unexpected_error as _is_unexpected_error
    from infi.vendata.vmware_powertools.snapshots.exceptions import SnapshotException
    from infi.vendata.vmware_powertools.vcenter.unmap import ValidationError
    from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.reclaim import BadEsxVersion
    from infi.vendata.vmware_powertools.mapping import NoHostsException
    if isinstance(error, (SnapshotException, ValidationError, BadEsxVersion, NoHostsException)):
        return False
    return _is_unexpected_error(error)

def json_response(func):
    @wraps(func)
    def callable(*args, **kwargs):
        if cherrypy.request.method in ('PUT', 'POST'):
            data = json.loads(cherrypy.request.body.read() or '{}')
            if isinstance(data, dict):
                kwargs.update(data)
            else:
                kwargs['data'] = data
        cherrypy.response.headers['Content-Type'] = "application/json"
        error_message = None
        return_value = None
        unexpected_error = False
        try:
            return_value = func(*args, **kwargs)
            success = True
        except (Exception, SystemExit) as error:
            from infinisdk.core.exceptions import APICommandFailed, APITransportFailure, SystemNotFoundException
            from httplib import BadStatusLine
            from pyVmomi import vim
            if isinstance(error, vim.MethodFault):
                error_message = "{}: {}".format(error.__class__.__name__.split(".")[-1], error.msg)
            elif isinstance(error, APICommandFailed):
                error_message = error.message
            elif isinstance(error, (SystemNotFoundException, APITransportFailure)):
                msg = "Failed to communicate with system. " \
                      "Please check IP connectivity to \"{}\" from the PowerTools server"
                error_message = msg.format(error.address)
            elif isinstance(error, BadStatusLine):
                error_message = "Error while communicating with vCenter or other remote host."
            else:
                error_message = unicode(error)
            if len(error_message.strip()) == 0:
                error_message = error.__class__.__name__
            unexpected_error = is_unexpected_error(error)
            msg = "Caught an exception on json response on function call {} with args {!r}, kwargs {!r}"
            logger.exception(msg.format(func.__name__, args, kwargs))
            success = False
        return json.dumps(dict(success=success, return_value=return_value, error_message=error_message,
            unexpected_error=unexpected_error))
    callable.exposed = True
    return callable

from .plugin import Plugin

class WebApplication(object):
    def __init__(self):
        super(WebApplication, self).__init__()

    def get_vcenter(self):
        from ..vcenter import vCenter
        return vCenter

    @cherrypy.expose
    def index(self):
        pass

    plugin = Plugin()

    @classmethod
    def get_wsgi_application(cls):
        from .error_pages import error_page
        from .plugin.resource import set_vmware_vcenter_client_request_hook
        from ..logging import LOG_DIRPATH
        set_vmware_vcenter_client_request_hook()
        logging_configuration = {'log.access_file': '',
                                 'log.error_file': '',
                                 'log.screen': False,
                                 'error_page.404': os.path.join(POWERTOOLS_STATIC_PATH, 'html', '404.html'),
                                 'error_page.default': error_page,
                                }
        cherrypy.config.update(logging_configuration)
        configuration = {'/plugin/views': {'tools.vmware_vcenter_client_request.on': True,
                                           'tools.assert_browser.on': True,
                                           'tools.redirect_vwc.on': True,
                                           'tools.assert_ready.on': True},
                         '/plugin/logs': {'tools.vmware_vcenter_client_request.on': True,
                                          'tools.redirect_vwc.on': True},
                         '/plugin/debug': {'tools.assert_ready.on': True},
                         '/plugin/redirect': {'tools.assert_browser.on': True,
                                              'tools.redirect_vwc.on': True,
                                              'tools.assert_ready.on': True},
                         '/favicon.ico': {'tools.staticfile.on': True,
                                          'tools.staticfile.filename': FAVICON_PATH},
                         '/static': {'tools.staticdir.on': True,
                                     'tools.redirect_vwc.on': True,
                                     'tools.staticdir.dir': POWERTOOLS_STATIC_PATH,
                                     'tools.staticdir.content_types': {'otf': 'application/x-font-otf'}},
                         '/static/vmware': {'tools.staticdir.on': True,
                                            'tools.redirect_vwc.on': True,
                                            'tools.staticdir.dir': STATIC_PATH},
                         '/vwc/vwc.zip': {'tools.staticfile.on': True,
                                          'tools.staticfile.filename': VWC_STATIC_PATH},
                         '/logs': {'tools.staticdir.on': True,
                                   'tools.response_headers.on': True,
                                   'tools.response_headers.headers': [('Content-Disposition', 'attachment')],
                                   'tools.staticdir.content_types': {'log': 'text/plain'},
                                   'tools.staticdir.dir': LOG_DIRPATH}}

        app = cherrypy.tree.mount(cls(), '/', configuration)
        return app

    @classmethod
    def clear_cache(cls):
        cherrypy.tree.apps[''].root.plugin.views.host.registration.clear_cache()
        cherrypy.tree.apps[''].root.plugin.views.cluster.registration.clear_cache()
        cherrypy.tree.apps[''].root.plugin.views.host.datastores.clear_cache()
