import cherrypy
import logging

from .resource import PluginResource
from .info import Info
from .views import Views
from .debug import Debug
from .redirect import Redirect
from .warmup import Warmup

logger = logging.getLogger(__name__)


def get_powertools_logs():
    from infi.vendata.vmware_powertools.logging import LOG_DIRPATH
    from infi.logs_collector.collectables import Directory
    from os import path
    from infi.vendata.vmware_powertools.scripts.greeter import ERROR_LOGFILE
    return [Directory(LOG_DIRPATH),
            Directory(path.dirname(ERROR_LOGFILE), path.basename(ERROR_LOGFILE) + '.*')]


def get_items():
    from infi.logs_collector.collectables import Command, Directory
    return [Command("free", ["-m"]),
            Command("infinivmware", ["--version"], prefix="infinivmware_version"),
            Command("ifconfig", ["-a"]),
            Command("df", ["-h"]),
            Directory("/var/log", recursive=True),
            Directory("/opt/vmware/var/log", recursive=True)]


def run(timestamp, delta, optional_archive_path=None, silent=False, interactive=False, only_infinihost_logs=False):
    from infi.logs_collector.scripts import parse_datestring, parse_deltastring
    timestamp = parse_datestring(timestamp)
    delta = parse_deltastring(delta)

    from infi.logs_collector import run
    items = get_powertools_logs()
    if not only_infinihost_logs:
        items += get_items()
    return run("infinihost", items, timestamp, delta, optional_archive_path,
               parent_dir_name="infinihost-diagnostics", silent=silent, interactive=interactive)


class Log(PluginResource):
    def _get_logs_content(self):
        from ...logging import LOG_DIRPATH
        from os import path
        logs_file = path.join(LOG_DIRPATH, "server.log")
        if path.exists(logs_file):
            with open(logs_file) as fd:
                return fd.read()
        else:
            return ''

    @cherrypy.expose
    def index(self, **kwargs):
        from re import MULTILINE, sub, split
        from time import time
        log_contents = self._get_logs_content()
        # split content by finding dates at beginning of lines
        log_entries = sub(r"^(\d+-\d+-\d+)", "\x00" + r"\1", log_contents, flags=MULTILINE).split("\x00")
        # limit number of entries
        log_entries = log_entries[:5000]
        # split entries to columns
        log_entries = [split(r"\s+", entry.rstrip(), 5, MULTILINE) for entry in log_entries if entry]
        # make sure there are enough columns in each entry
        log_entries = [entry for entry in log_entries if len(entry) > 5]
        timestamp = time()
        return self.template_lookup.get_template("logs.mako").render(log_entries=log_entries, timestamp=timestamp)

    @cherrypy.expose
    def serverlog(self, **kwargs):
        from time import strftime
        filename = strftime("server_%Y_%d_%m_%H_%M_%S.log")
        cherrypy.response.headers['Content-Type'] = "text/plain"
        cherrypy.response.headers['Content-Disposition'] = 'attachment; filename="{}"'.format(filename)
        return self._get_logs_content()

    @cherrypy.expose
    def collect(self, **kwargs):
        from cherrypy.lib.static import serve_file
        from time import time
        if "download" in kwargs:
            rc, filepath = run('now', '1w', silent=True)
            return serve_file(filepath, "application/x-gzip", "attachment")
        else:
            return self.template_lookup.get_template('collect_logs.mako').render(timestamp=time())


class Plugin(PluginResource):
    info = Info()
    logs = Log()
    views = Views()
    debug = Debug()
    redirect = Redirect()
    warmup = Warmup()
