import cherrypy
import logging
import cjson
from munch import Munch
from ...resource import ExtensionResource
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.vcenter.shortcuts import object_to_moref
from infi.vendata.vmware_powertools.mapping import good_cluster_hosts

logger = logging.getLogger(__name__)

from .datastores import ClusterDatastores, ClusterDatastoresTable, ClusterBoxVolumes
from .readiness import Readiness
from .registration import ClusterRegistration
from .operations import Operations

def get_infiniboxes_per_host(result_dict, host):
    host_get_infiniboxes = cherrypy.tree.apps[''].root.plugin.views.host._get_infiniboxes
    result_dict[host] = host_get_infiniboxes(object_to_moref(host))

def get_infiniboxes(cluster_compute_resource):
    from infi.vendata.vmware_powertools.utils import spawn, joinall
    infiniboxes_per_host = {}
    joinall([spawn(get_infiniboxes_per_host, infiniboxes_per_host, host) for
             host in cluster_compute_resource.host], raise_error=True)
    serials_per_host = [set([infinibox.serial for infinibox in value])
                        for value in infiniboxes_per_host.values()]
    if serials_per_host == []:
        return []
    intersecting_serials = set.intersection(*serials_per_host)
    return [infinibox for infinibox in infiniboxes_per_host.values()[0]
            if infinibox.serial in intersecting_serials]

class Cluster(ExtensionResource):
    EXTENSION_POINT = "InventoryView.Cluster"

    @cherrypy.expose
    def index(self, moref, **kwargs):
        return self.get_response(moref)

    def get_response(self, moref):
        from time import time
        from infi.vendata.vmware_powertools.__version__ import __version__
        encode = getattr(cjson, "encode")
        url = self.get_url_from_request()
        sources = self._get_sources(url)
        columns = self._get_columns()
        infiniboxes = self._get_infiniboxes(moref)
        clustername = self._get_clustername(moref)
        timestamp = time()
        kwargs = dict(sources=encode(sources), infiniboxes=infiniboxes, table_columns=columns,
                      clustername=clustername, timestamp=timestamp, powertools_version=__version__)
        return self.template_lookup.get_template('cluster.mako').render(**kwargs)

    def _get_cluster(self, moref):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object
        return moref_to_object(moref)

    def _get_clustername(self, moref):
        cluster = self._get_cluster(moref)
        return cluster.name

    def _get_infiniboxes(self, moref):
        cluster = self._get_cluster(moref)
        return get_infiniboxes(cluster)

    def _get_sources(self, url):
        url = url.replace("/redirect", "/views")
        url_format = url.replace("/cluster/?", "/cluster/{uri}/?")
        sources = Munch(
            status=url_format.format(uri="readiness"),
            refresh=url_format.format(uri="refresh"),
            registration=url_format.format(uri="registration"),
            datastore_table=url_format.format(uri="datastores"),
            box_volumes=url_format.format(uri="box_volumes"),
        )
        sources.operations = Munch(
            unmap=url_format.format(uri="operations/unmap"),
            delete_vol=url_format.format(uri="operations/delete_vol"),
            get_provision_tasks=url_format.format(uri="operations/provisioning/get_provision_tasks"),
            provision_params=url_format.format(uri="operations/provisioning/provision_params"),
            provision_validation=url_format.format(uri="operations/provisioning/validate_provision_params"),
            run_provision_task=url_format.format(uri="operations/provisioning/run_provision_task"),
            map_existing_volume=url_format.format(uri="operations/map_existing"),
            query_mount_after_map=url_format.format(uri="operations/query_mount_after_map"),
            create_filesystem=url_format.format(uri="operations/create_filesystem"),
            resize_validation=url_format.format(uri="operations/resize/resize_validation"),
            get_resize_tasks=url_format.format(uri="operations/resize/get_resize_tasks"),
            run_resize_task=url_format.format(uri="operations/resize/run_resize_task"),
            rename=url_format.format(uri="operations/rename"),
            reclaim=url_format.format(uri="operations/reclaim"),
            rescan=url_format.format(uri="operations/rescan"),
        )
        sources.auth = Munch(
            has_credentials=url.replace("/cluster/?", "/auth/has_credentials/?"),
            set_credentials=url.replace("/cluster/?", "/auth/set_credentials/?"),
            reset_credentials=url.replace("/cluster/?", "/auth/reset_credentials/?"),
            get_current_username=url.replace("/cluster/?", "/auth/get_current_username/?")
        )
        return sources

    def _get_columns(self):
        return ClusterDatastoresTable.COLUMN_NAMES

    @json_response
    def refresh(self, moref, **kwargs):
        cluster = self._get_cluster(moref)
        for host in good_cluster_hosts(cluster):
            host_moref = object_to_moref(host)
            cherrypy.tree.apps[''].root.plugin.views.host.refresh(host_moref)

        for infinibox in self._get_infiniboxes(moref):
            self.registration.clear_cached_entry(self.registration.GET, moref, infinibox['serial'])

    registration = ClusterRegistration()
    readiness = Readiness()
    datastores = ClusterDatastores()
    operations = Operations()
    box_volumes = ClusterBoxVolumes()
