import cherrypy
from infi.vendata.vmware_powertools.webapp.plugin.resource import PluginResource
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.rename import HostRenameVolume
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.reclaim import reclaim, BadEsxVersion, esx_supports_reclaim
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.utils import fix_serial, get_system_serial
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref
from infi.vendata.vmware_powertools.mapping import get_all_mappings_in_cluster, good_cluster_hosts
from .map import ClusterMapVolume
from .unmap import ClusterUnmapVolume
from .resize import Resize
from .provision import Provisioning

def cluster_datastores_changed(esx_cluster, system_serial):
    # Called when a datastore was added to the cluster or removed from it.
    # Clears the cache of the cluster's hosts' datastores.
    host_datastores = cherrypy.tree.apps[''].root.plugin.views.host.datastores
    for esx_host in good_cluster_hosts(esx_cluster):
        host_moref = object_to_moref(esx_host)
        host_datastores.clear_cached_entry(host_datastores.get_row_data, host_moref, system_serial)


class Operations(PluginResource):
    @json_response
    def map_existing(self, moref, serial, volume_id, **kwargs):
        esx_cluster = moref_to_object(moref)
        ClusterMapVolume().run(esx_cluster, serial, volume_id)
        cluster_datastores_changed(esx_cluster, int(serial))

    @json_response
    def query_mount_after_map(self, serial, volume_id, **kwargs):
        return dict(has_fs=False)

    @json_response
    def unmap(self, moref, serial, **kwargs):  # pragma: no cover
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        ClusterUnmapVolume(esx_cluster, proper_serial).unmap()
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))

    @json_response
    def delete_vol(self, moref, serial, **kwargs):  # pragma: no cover
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        ClusterUnmapVolume(esx_cluster, proper_serial).delete()
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))

    @json_response
    def rescan(self, moref, **kwargs):
        from infi.vendata.vmware_powertools.utils import refresh_storage_model_on_cluster
        refresh_storage_model_on_cluster(moref_to_object(moref), raise_error=True, do_rescan=True, do_refresh=True)

    @json_response
    def rename(self, moref, serial, name, rename_datastore, rename_device, **kwargs):
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        host_mappings = get_all_mappings_in_cluster(esx_cluster, proper_serial)
        host = host_mappings[0]
        HostRenameVolume().rename(host, proper_serial, name, rename_datastore, rename_device)
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))

    @json_response
    def reclaim(self, moref, serial, name, **kwargs):
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        host_mappings = get_all_mappings_in_cluster(esx_cluster, proper_serial)
        host_mappings = [host for host in host_mappings if esx_supports_reclaim(host)]
        if len(host_mappings) == 0:
            raise BadEsxVersion("No ESX host found in cluster that supports the Reclaim Space operation")
        host = host_mappings[0]
        result = reclaim(host, proper_serial, name)
        cluster_datastores_changed(esx_cluster, get_system_serial(serial))
        return dict(reclaimed=result)

    @json_response
    def create_filesystem(self, moref, serial, **kwargs):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.utils import create_vmfs_datastore_by_serial, mount_vmfs_datastore_by_serial
        esx_cluster = moref_to_object(moref)
        proper_serial = fix_serial(serial)
        first = True
        for esx_host in good_cluster_hosts(esx_cluster, maintenance_hosts=False):
            func = create_vmfs_datastore_by_serial if first else mount_vmfs_datastore_by_serial
            func(esx_host, proper_serial)
            first = False

    resize = Resize()
    provisioning = Provisioning()
