import logging
from infi.vendata.powertools.utils import queries
from infi.vendata.vmware_powertools.webapp.plugin.resource import PluginResource
from infi.vendata.vmware_powertools.utils import spawn, joinall
from infi.vendata.vmware_powertools.mapping import good_cluster_hosts
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.provision import ProvisionMixin, convert_size
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object
from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.utils import TaskMixin

logger = logging.getLogger(__name__)


class ClusterProvisionVolumeTask(ProvisionMixin, TaskMixin, object):
    def __init__(self, esx_cluster, system_serial, pool_name, volume_name, volume_size, thin, do_filesystem, set_device_name, volume_count=1):
        super(ClusterProvisionVolumeTask, self).__init__()
        self._esx_cluster = esx_cluster
        self._esx_object = esx_cluster
        self._system_serial = system_serial
        self._pool_name = pool_name
        self._volume_name = volume_name
        self._volume_size = volume_size
        self._volume_count = volume_count
        self._thin = thin
        self._do_filesystem = do_filesystem
        self._set_device_name = set_device_name
        self._task_name = "Volume provisioning"
        self._hosts = good_cluster_hosts(self._esx_cluster, maintenance_hosts=False)
        self._cluster = None
        self._volumes = []

    def _preparation(self):
        from random import choice
        # pick primary provisioning host
        self._esx_host = choice(self._hosts)
        logger.info("Picked ESX host '{}' as main provisioing host for cluster".format(self._esx_host.name))

        super(ClusterProvisionVolumeTask, self)._preparation()

    def _volume_map(self):
        self._task.set_description("Mapping volume{} to cluster" + self._suffix)
        self._cluster = self._system.host_clusters.safe_get(id=self._system_info.get_cluster_id())
        if self._cluster is None:
            raise queries.ObjectNotFound("Could not find requested cluster object in InfiniBox")
        [self._cluster.map_volume(volume) for volume in self._volumes]

    def _mount_on_other_hosts(self):
        def op(host):
            self._vmfs_mount(host)
        other_hosts = [host for host in self._hosts if host._moId != self._esx_host._moId]
        joinall([spawn(op, host) for host in other_hosts], raise_error=True)
        self._task.set_description("Waiting for mounts to appear")
        self._rescan_on_all_hosts(set_description=False)

    def _rescan_on_all_hosts(self, set_description=True):
        if set_description:
            self._task.set_description("Waiting for volume{} to appear".format(self._suffix))
        def op(host):
            self._wait_for_volume(host)
        joinall([spawn(op, host) for host in self._hosts], raise_error=True)

    def _finish(self):
        from . import cluster_datastores_changed
        joinall([spawn(self._refresh_storagemodel, host) for host in self._hosts], raise_error=True)
        cluster_datastores_changed(self._esx_cluster, self._system_serial)
        return super(ClusterProvisionVolumeTask, self)._finish()

    def _get_steps(self):
        steps = [
            self._preparation,
            self._volume_create,
            self._volume_map,
            self._rescan_on_all_hosts
        ]
        if self._do_filesystem:
            steps.append(self._vmfs_create)
            steps.append(self._mount_on_other_hosts)
        steps.append(self._finish)
        return steps


class Provisioning(PluginResource):
    @json_response
    def provision_params(self, moref, serial, **kwargs):
        from ...host.operations.provision import get_default_provision_params
        esx_cluster = moref_to_object(moref)
        return get_default_provision_params(good_cluster_hosts(esx_cluster, maintenance_hosts=False)[0], serial)    # we use the first host, there should be no difference

    @json_response
    def get_provision_tasks(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        suffix = "s" if int(count) > 1 else ""
        series = " series" if int(count) > 1 else ""
        series_count = " ({} volumes)".format(count) if int(count) > 1 else ""
        ret = [
                'Connecting to InfiniBox',
                'Creating the new {} volume{} "{}"{}'.format(volume_size, series, volume_name, series_count),
                'Mapping volume{} to cluster'.format(suffix),
                'Waiting for volume discovery, this might take a while'
        ]
        if do_filesystem:
            ret.append('Creating filesystem')
            ret.append('Mounting volume on all remaining hosts in cluster')
        ret.append('Performing housekeeping')
        return ret

    @json_response
    def validate_provision_params(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        from ...host.operations.provision import validate_provision_params
        esx_cluster = moref_to_object(moref)
        proper_serial = int(serial)
        host = good_cluster_hosts(esx_cluster, maintenance_hosts=False)[0]
        validate_provision_params(host, proper_serial, pool_name, volume_name, volume_size, count, do_filesystem)

    @json_response
    def run_provision_task(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        esx_cluster = moref_to_object(moref)
        task_key = (ClusterProvisionVolumeTask, moref)
        proper_serial = int(serial)
        proper_size = convert_size(volume_size)
        volume_count = int(count)
        task_object_creator = lambda: ClusterProvisionVolumeTask(esx_cluster, proper_serial, pool_name, volume_name, proper_size, thin, do_filesystem, set_device_name, volume_count)
        return self.next_task_step(task_key, task_object_creator)
