import cherrypy
import logging
import cjson
from munch import Munch
from ...resource import ExtensionResource
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.powertools.utils.versioncheck import is_supported

logger = logging.getLogger(__name__)

from .readiness import Readiness
from .datastores import HostDatastores, HostDatastoresTable, BoxVolumes
from .registration import HostRegistration
from .operations import Operations

def get_infiniboxes(esx_host):
    from infi.vendata.vmware_powertools.patches import with_host
    from infi.vendata.vmware_powertools.utils.auth import auth
    infiniboxes = []
    with with_host(esx_host):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices_and_controllers__mapped_to_lun0
        try:
            devices = get_infinidat_block_devices_and_controllers__mapped_to_lun0()
        except:
            logger.exception("failed to get infiniboxes for esx host {}, returning an empty list; user should refresh".format(esx_host))
            return []
        for device in devices:
            vendor = device.get_vendor()
            serial = vendor.get_system_serial()
            credentials = auth.CREDENTIALS_STORE.get_existing_credentials(serial)
            username = None if credentials is None else credentials.get_username()
            url = "https://{}".format(vendor.get_management_address())
            version = vendor.get_system_version()
            infiniboxes.append(Munch(name=vendor.get_system_name(),
                                     serial=serial,
                                     gui=url,
                                     version=version,
                                     username=username,
                                     supported=is_supported(version)))
    return infiniboxes

class Host(ExtensionResource):
    EXTENSION_POINT = "InventoryView.HostSystem"

    @cherrypy.expose
    def index(self, moref, **kwargs):
        return self.get_response(moref)

    def get_response(self, moref):
        from time import time
        from infi.vendata.vmware_powertools.__version__ import __version__
        from infi.vendata.vmware_powertools.mapping import is_host_connected, is_host_in_maintenance_mode
        encode = getattr(cjson, "encode")
        url = self.get_url_from_request()
        sources = self._get_sources(url)
        columns = self._get_columns()
        infiniboxes = self._get_infiniboxes(moref)
        timestamp = time()
        host = self._get_host(moref)
        hostname = host.summary.config.name
        host_disconnected = not is_host_connected(host)
        host_in_maintenance = is_host_in_maintenance_mode(host)
        kwargs = dict(sources=encode(sources), infiniboxes=infiniboxes, table_columns=columns,
                      hostname=hostname, timestamp=timestamp, powertools_version=__version__,
                      host_disconnected=host_disconnected, host_in_maintenance=host_in_maintenance)
        return self.template_lookup.get_template('host.mako').render(**kwargs)

    def _get_host(self, moref):
        from infi.vendata.vmware_powertools.vcenter import vCenter
        client = vCenter.get_extension_client()
        esx_host = client.get_managed_object_by_reference(moref)
        return esx_host

    @ExtensionResource.cached_method
    def _get_infiniboxes(self, moref):
        host = self._get_host(moref)
        return get_infiniboxes(host)

    def _get_sources(self, url):
        url = url.replace("/redirect", "/views")
        url_format = url.replace("/host/?", "/host/{uri}/?")
        sources = Munch(
            status=url_format.format(uri="readiness"),
            refresh=url_format.format(uri="refresh"),
            registration=url_format.format(uri="registration"),
            datastore_table=url_format.format(uri="datastores"),
            box_volumes=url_format.format(uri="box_volumes"),
        )
        sources.operations = Munch(
            unmap=url_format.format(uri="operations/unmap"),
            delete_vol=url_format.format(uri="operations/delete_vol"),
            get_provision_tasks=url_format.format(uri="operations/provisioning/get_provision_tasks"),
            provision_params=url_format.format(uri="operations/provisioning/provision_params"),
            provision_validation=url_format.format(uri="operations/provisioning/validate_provision_params"),
            run_provision_task=url_format.format(uri="operations/provisioning/run_provision_task"),
            map_existing_volume=url_format.format(uri="operations/map_existing"),
            query_mount_after_map=url_format.format(uri="operations/query_mount_after_map"),
            create_filesystem=url_format.format(uri="operations/create_filesystem"),
            resize_validation=url_format.format(uri="operations/resize/resize_validation"),
            get_resize_tasks=url_format.format(uri="operations/resize/get_resize_tasks"),
            run_resize_task=url_format.format(uri="operations/resize/run_resize_task"),
            rename=url_format.format(uri="operations/rename"),
            reclaim=url_format.format(uri="operations/reclaim"),
            rescan=url_format.format(uri="operations/rescan"),
            update_metadata=url_format.format(uri="operations/update_metadata"),
        )
        sources.auth = Munch(
            has_credentials=url.replace("/host/?", "/auth/has_credentials/?"),
            set_credentials=url.replace("/host/?", "/auth/set_credentials/?"),
            reset_credentials=url.replace("/host/?", "/auth/reset_credentials/?"),
            get_current_username=url.replace("/host/?", "/auth/get_current_username/?")
        )
        return sources

    def _get_columns(self):
        return HostDatastoresTable.COLUMN_NAMES

    @json_response
    def refresh(self, moref, **kwargs):
        from infi.vendata.vmware_powertools.patches import with_host
        host = self._get_host(moref)
        with with_host(host):
            from infi.storagemodel import get_storage_model
            get_storage_model().refresh()

        for infinibox in self._get_infiniboxes(moref):
            self.datastores.clear_cached_entry(self.datastores.get_row_data, moref, infinibox['serial'])
            self.registration.clear_cached_entry(self.registration.GET, moref, infinibox['serial'])
        self.clear_cached_entry(self._get_infiniboxes, moref)

    readiness = Readiness()
    registration = HostRegistration()
    datastores = HostDatastores()
    operations = Operations()
    box_volumes = BoxVolumes()
