import cherrypy
import logging
from munch import Munch
from ....resource import PluginResource
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref
from infi.vendata.vmware_powertools.logging import log_timing

logger = logging.getLogger(__name__)

NO_DATASTORE = Munch(summary=Munch(name="<No Datastore>"))


class HostDatastoresTable(object):
    COLUMN_NAMES = ["Datastore Name", "Volume Name", "Size", "# Paths", "Lun", "Device Name", "Actions"]

    def __init__(self, esx_host):
        self.esx_host = esx_host

    def is_device_belongs_to_this_system(self, device, system_serial):
        return device.get_vendor().get_system_serial() == system_serial

    def _get_datastore_element(self, datastore):
        if datastore is NO_DATASTORE:
            return NO_DATASTORE.summary.name, None
        datastore_name = datastore.summary.name
        datastore_ref = object_to_moref(datastore)
        return datastore_name, datastore_ref

    def _get_volume_element(self, infinidat_device):
        vendor = infinidat_device.get_vendor()
        base_url = "https://{}".format(vendor.get_management_address())
        url = "{}/#/volumes/{}".format(base_url, vendor.get_volume_id())
        return vendor.get_volume_name(), url

    def _is_mounted(self, datastore):
        if datastore is NO_DATASTORE:
            return True     # don't add 'unmounted' text and style - the info already says "no datastore"
        matches = [datastore_host.mountInfo.mounted for datastore_host in datastore.host
                   if datastore_host.key == self.esx_host]
        return len(matches) == 1 and matches[0]

    def build_row(self, datastore, infinidat_device):
        from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations import esx_supports_reclaim
        datastore_name, datastore_ref = self._get_datastore_element(datastore)
        volume_name, volume_url = self._get_volume_element(infinidat_device)
        thin = infinidat_device.is_thin_provisioned()
        reclaim_available = esx_supports_reclaim(self.esx_host) and thin and datastore_ref is not None
        return dict(datastore_name=datastore_name,
                    datastore_ref=datastore_ref,
                    volume_name=volume_name,
                    volume_url=volume_url,
                    volume_size=infinidat_device.get_size_in_bytes(),
                    paths=len(infinidat_device.get_paths()),
                    lun=infinidat_device.get_paths()[0].get_hctl().get_lun(),
                    device_name=infinidat_device.get_display_name(),
                    serial=infinidat_device.get_scsi_serial_number(),
                    thin=thin,
                    mounted=self._is_mounted(datastore),
                    reclaim_available=reclaim_available)

    def _iter_infinidat_datastores(self):
        from infi.vendata.vmware_powertools.vcenter.shortcuts import iter_infinidat_datastores
        return iter_infinidat_datastores(self.esx_host)

    def build_rows(self, system_serial):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
        all_infinibox_devices = set([device for device in get_infinidat_block_devices()
                                     if self.is_device_belongs_to_this_system(device, system_serial)])
        rows = []
        for datastore, infinidat_devices in self._iter_infinidat_datastores():
            for device in infinidat_devices:
                if device in all_infinibox_devices:
                    rows.append(self.build_row(datastore, device))
                    all_infinibox_devices.remove(device)
        for device in all_infinibox_devices:
            rows.append(self.build_row(NO_DATASTORE, device))
        return rows

    def get_row_data(self, system_serial):
        from infi.vendata.vmware_powertools.patches import with_host
        with with_host(self.esx_host):
            return self.build_rows(system_serial)

class DatastoresResource(PluginResource):
    @json_response
    def index(self, moref, serial, **kwargs):
        return self.get_response(moref, int(serial))

    def get_response(self, moref, serial):
        return self.get_row_data(moref, serial)

    @log_timing
    def get_row_data(self, moref, serial):
        table = self._get_table_builder(moref)
        return table.get_row_data(serial)

    def _get_table_builder(self, moref):
        esx_host = moref_to_object(moref)
        return HostDatastoresTable(esx_host)


class HostDatastores(DatastoresResource):
    get_row_data = PluginResource.cached_method(DatastoresResource.get_row_data)

    def clear_cache_by_datastore(self, datastore_serial):
        """ Clear the cached row data for all hosts that are mapped to the given datastore """
        from infi.vendata.vmware_powertools.mapping import get_all_host_mappings
        from infi.vendata.vmware_powertools.webapp.plugin.views.host.operations.utils import get_system_serial
        system_serial = get_system_serial(datastore_serial)
        for host in get_all_host_mappings(datastore_serial):
            self.clear_cached_entry(self.get_row_data, object_to_moref(host), system_serial)


class BoxVolumes(object):
    @cherrypy.expose
    def index(self, moref, serial, **kwargs):
        from infi.vendata.vmware_powertools.patches import with_host
        from infi.vendata.powertools.gui.webapp.plugin.views.host.volumes import BoxVolumes
        esx_host = moref_to_object(moref)
        with with_host(esx_host):
            return BoxVolumes().index(serial, show_mapped=True, **kwargs)
