from infi.vendata.vmware_powertools.webapp.plugin.resource import PluginResource
from infi.vendata.vmware_powertools.webapp import json_response
from .utils import update_device_name, create_vmfs_datastore_with_metadata, host_datastores_changed, TaskMixin
from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import ProvisionValidationException
from infi.vendata.powertools.utils import round_to_block_size
from infi.vendata.powertools.utils import queries
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object, object_to_moref
from infi.vendata.vmware_powertools.patches import with_host
from infi.vendata.vmware_powertools.utils import spawn, joinall
from capacity import byte


def get_default_provision_params(esx_host, serial):
    from infi.vendata.vmware_powertools.vcenter.map import HostSystemByController
    import uuid
    serial = int(serial)
    infinisdk_system = HostSystemByController(esx_host, serial).get_system()
    pools = [dict(name=pool.get_name(),
                  size=pool.get_virtual_capacity() // byte,
                  free=pool.get_free_virtual_capacity() // byte) for pool in queries.get_pools(infinisdk_system)]
    pools.sort(key=lambda pool: pool["name"].lower())
    return dict(box_serial=serial,
                filesystems=[dict(name="VMFS")],
                selected_filesystem="VMFS",
                suggested_volume_name="volume_{}".format(str(uuid.uuid1()).replace("-", "")),
                suggested_mountpoint="",
                pools=pools)

def validate_provision_params(esx_host, box_serial, pool_name, volume_name, volume_size, count, with_filesystem):
    from infi.vendata.powertools.cli.autodetect import get_available_pools
    from infi.vendata.powertools.cli.validate import validate_volume_name
    if not count.isdigit():
        raise ProvisionValidationException("Volume count is invalid")
    if int(count) > 20 or int(count) <= 0:
        raise ProvisionValidationException("Volume count must be between 1 and 20")
    try:
        int(box_serial)
    except ValueError:
        raise ProvisionValidationException("Box serial is invalid (expected a number)")
    with with_host(esx_host):
        if pool_name not in get_available_pools(box_serial):
            raise ProvisionValidationException("Unavailable pool")
        validation_result = validate_volume_name(volume_name, box_serial)
        if validation_result:
            raise ProvisionValidationException(validation_result)
        validate_size(box_serial, pool_name, volume_size, count, with_filesystem)

def validate_filesystem_size(volume_size):
    from capacity import TiB
    if volume_size > 64*TiB:
        return "Maximum size of VMFS volume exceeded. Maximum size is 64TB"

def validate_size(box_serial, pool_name, volume_size, count, with_filesystem):
    from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import validate_size as powertools_validate_size
    from infi.vendata.powertools.cli.validate import validate_pool_size
    if isinstance(volume_size, basestring):
        powertools_validate_size(volume_size)
        volume_size = convert_size(volume_size)
    validation_result = validate_pool_size(box_serial, pool_name, volume_size, int(count))
    if validation_result:
        raise ProvisionValidationException(validation_result)
    if with_filesystem:
        validation_result = validate_filesystem_size(volume_size)
        if validation_result:
            raise ProvisionValidationException(validation_result)

def convert_size(volume_size):
    from infi.vendata.powertools.cli.validate import convert_size as powertools_convert_size
    volume_size = volume_size.replace("GB", "GiB").replace("TB", "TiB")
    return powertools_convert_size(volume_size)


class ProvisionMixin(object):
    def _preparation(self):
        from infi.vendata.vmware_powertools.utils import get_system_object_by_device
        self._suffix = "s" if self._volume_count > 1 else ""
        self._task.set_description("Connecting to InfiniBox")
        with with_host(self._esx_host):
            system_controller = queries.get_system_controller_by_serial(self._system_serial)
            self._system_info = system_controller.get_vendor()
            self._system = get_system_object_by_device(system_controller)

    def _volume_create(self):
        self._task.set_description("Creating volume" + self._suffix)
        self._pool = self._system.pools.safe_get(name=self._pool_name)
        if self._pool is None:
            raise queries.ObjectNotFound("Could not find pool with name {}".format(self._pool_name))
        size = round_to_block_size(self._volume_size // byte) * byte
        provisioning = "thin" if self._thin else "thick"
        for i in range(self._volume_count):
            name = "{}{:03d}".format(self._volume_name, i+1) if self._volume_count > 1 else self._volume_name
            volume = self._system.volumes.create(pool=self._pool, name=name, size=size, provisioning=provisioning)
            self._volumes.append(volume)

    def _volume_map(self):
        self._task.set_description("Mapping volume{} to host".format(self._suffix))
        self._host = self._system.hosts.safe_get(id=self._system_info.get_host_id())
        if self._host is None:
            raise queries.ObjectNotFound("Could not find requested host object in InfiniBox")
        [self._host.map_volume(volume) for volume in self._volumes]

    def _get_block_device_by_volume_id(self, volume_id):
        from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
        return [device for device in get_infinidat_block_devices()
                if device.get_vendor().get_system_serial() == self._system_serial and
                   device.get_vendor().get_volume_id() == volume_id][0]

    def _wait_for_volume(self, esx_host=None):
        from infi.storagemodel.predicates import PredicateList
        from infi.storagemodel.vendor.infinidat.predicates import InfinidatVolumeExists
        from infi.storagemodel import get_storage_model
        if esx_host is None:
            self._task.set_description("Waiting for volume{} to appear".format(self._suffix))
        predicates = [InfinidatVolumeExists(self._system_serial, volume.get_id()) for volume in self._volumes]
        with with_host(esx_host or self._esx_host):
            get_storage_model().rescan_and_wait_for(PredicateList(predicates))
            self._block_devices = [self._get_block_device_by_volume_id(volume.get_id()) for volume in self._volumes]
            if self._set_device_name:
                joinall([spawn(update_device_name, esx_host or self._esx_host, block_device)
                         for block_device in self._block_devices], raise_error=True)

    def _vmfs_create(self):
        self._task.set_description("Creating VMFS")
        with with_host(self._esx_host):
            joinall([spawn(create_vmfs_datastore_with_metadata, self._esx_host, block_device, volume)
                     for block_device, volume in zip(self._block_devices, self._volumes)], raise_error=True)

    def _vmfs_mount(self, esx_host=None):
        from .utils import mount_vmfs_datastore
        joinall([spawn(mount_vmfs_datastore, esx_host or self._esx_host, block_device)
                 for block_device in self._block_devices], raise_error=True)

    def _refresh_storagemodel(self, esx_host=None):
        with with_host(esx_host or self._esx_host):
            from infi.storagemodel import get_storage_model
            get_storage_model().refresh()


class HostProvisionVolumeTask(ProvisionMixin, TaskMixin, object):
    def __init__(self, esx_host, system_serial, pool_name, volume_name, volume_size, thin, do_filesystem, set_device_name, volume_count=1):
        super(HostProvisionVolumeTask, self).__init__()
        self._esx_host = esx_host
        self._esx_object = esx_host
        self._system_serial = system_serial
        self._pool_name = pool_name
        self._volume_name = volume_name
        self._volume_size = volume_size
        self._volume_count = volume_count
        self._thin = thin
        self._do_filesystem = do_filesystem
        self._set_device_name = set_device_name
        self._task_name = "Volume provisioning"
        self._host = None
        self._volumes = []

    def _finish(self):
        self._refresh_storagemodel()
        moref = object_to_moref(self._esx_host)
        host_datastores_changed(moref, self._system_serial)
        return super(HostProvisionVolumeTask, self)._finish()

    def _get_steps(self):
        steps = [
            self._preparation,
            self._volume_create,
            self._volume_map,
            self._wait_for_volume
        ]
        if self._do_filesystem:
            steps.append(self._vmfs_create)
        steps.append(self._finish)
        return steps


class Provisioning(PluginResource):
    @json_response
    def provision_params(self, moref, serial, **kwargs):
        esx_host = moref_to_object(moref)
        return get_default_provision_params(esx_host, serial)

    @json_response
    def validate_provision_params(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        host = moref_to_object(moref)
        proper_serial = int(serial)
        validate_provision_params(host, proper_serial, pool_name, volume_name, volume_size, count, do_filesystem)

    @json_response
    def get_provision_tasks(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        suffix = "s" if int(count) > 1 else ""
        series = " series" if int(count) > 1 else ""
        series_count = " ({} volumes)".format(count) if int(count) > 1 else ""
        ret = [
                'Connecting to InfiniBox',
                'Creating the new {} volume{} "{}"{}'.format(volume_size, series, volume_name, series_count),
                'Mapping volume{} to host'.format(suffix),
                'Waiting for volume discovery, this might take a while'
        ]
        if do_filesystem:
            ret.append('Creating filesystem')
        ret.append('Performing housekeeping')
        return ret

    @json_response
    def run_provision_task(self, moref, serial, volume_name, pool_name, volume_size, thin, do_filesystem, set_device_name, count, **kwargs):
        host = moref_to_object(moref)
        task_key = (HostProvisionVolumeTask, moref)
        proper_serial = int(serial)
        proper_size = convert_size(volume_size)
        volume_count = int(count)
        task_object_creator = lambda: HostProvisionVolumeTask(host, proper_serial, pool_name, volume_name, proper_size, thin, do_filesystem, set_device_name, volume_count)
        return self.next_task_step(task_key, task_object_creator)
