from infi.vendata.vmware_powertools.webapp.plugin.resource import PluginResource
from infi.vendata.vmware_powertools.webapp import json_response
from infi.vendata.powertools.gui.webapp.plugin.views.host.operations import ProvisionValidationException
from .utils import get_datastore_by_serial_number, datastore_changed, TaskMixin
from infi.vendata.vmware_powertools.utils import get_infinidat_device_by_serial
from infi.vendata.vmware_powertools.utils import get_infinisdk_volume_by_device
from infi.vendata.vmware_powertools.patches import with_host
from infi.vendata.vmware_powertools.utils import spawn, joinall
from .utils import fix_serial
from infi.vendata.vmware_powertools.vcenter.shortcuts import moref_to_object
from .provision import convert_size
from infi.vendata.vmware_powertools.mapping import get_all_host_mappings
from capacity import byte, GiB


class HostResizeVolumeTask(TaskMixin, object):
    def __init__(self, esx_host, proper_serial, proper_size):
        super(HostResizeVolumeTask, self).__init__()
        self._esx_host = esx_host
        self._esx_object = esx_host
        self._proper_serial = proper_serial
        self._proper_size = proper_size
        self._original_size = None
        self._volume = None
        self._task_name = "Volume resize"

    def get_task_list(self):
        res = ['Resizing the volume',
               'Rescanning']
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        if datastore is not None:
            res.extend(['Resizing the filesystem'])
        res.extend([
                'Rescanning all related hosts',
                'Performing housekeeping'])
        return res

    def _resize_datastore(self):
        self._task.set_description("Resizing VMFS datastore")
        manager = self._esx_host.configManager.datastoreSystem
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        available_resize_options = manager.QueryVmfsDatastoreExpandOptions(datastore)
        # we can either resize by adding an extent, or by expanding an extend; we want to expand
        if not available_resize_options:
            raise ProvisionValidationException("No VMFS expand options")
        expand_spec = available_resize_options[0].spec
        manager.ExpandVmfsDatastore(datastore, expand_spec)

    def _size_changed_predicate(self):
        device = get_infinidat_device_by_serial(self._esx_host, self._proper_serial)
        return device.get_size_in_bytes() == (self._proper_size // byte)

    def _rescan(self):
        self._task.set_description("Rescanning")
        with with_host(self._esx_host):
            from infi.storagemodel import get_storage_model
            get_storage_model().rescan_and_wait_for(self._size_changed_predicate, do_rescan=False, do_refresh=True)
            datastore_changed(self._proper_serial)

    def _rescan_other_hosts(self):
        self._task.set_description("Rescanning all related hosts")
        # We need to rescan on all hosts that are mapped to the volume, not just this host.
        def rescan(esx_host):
            with with_host(esx_host):
                from infi.storagemodel import get_storage_model
                get_storage_model().rescan_and_wait_for(do_rescan=False, do_refresh=True)
                datastore_changed(self._proper_serial)
        greenlets = [spawn(rescan, esx_host) for esx_host in get_all_host_mappings(self._proper_serial)
                     if esx_host._moId != self._esx_host._moId]
        # We don't really wait for all the hosts and we don't add a predicate, because some hosts may have problems
        # and may take a lot of time to rescan. So we do best-effort...
        # We do wait a short time so that this step will appear in the task description - this way the rescans have
        # a context from the user's perspective.
        joinall(greenlets, timeout=3)

    def _get_device_and_volume(self):
        with with_host(self._esx_host):
            device = get_infinidat_device_by_serial(self._esx_host, self._proper_serial)
            volume = get_infinisdk_volume_by_device(device)
        return device, volume

    def assert_resize_available(self):
        from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
        from .provision import validate_filesystem_size
        VolumeUnmapBase(self._esx_host, self._proper_serial).assert_not_being_used_as_rdm()
        device, volume = self._get_device_and_volume()
        old_size = volume.get_size()
        new_size = self._proper_size
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        has_fs = datastore is not None
        if new_size < old_size:
            raise ProvisionValidationException("Cannot shrink volumes")
        if (new_size - old_size) > volume.get_pool().get_free_virtual_capacity():
            raise ProvisionValidationException("Not enough free space in pool")
        if has_fs:
            if new_size <= old_size + (0.3 * GiB):
                raise ProvisionValidationException("New size must be at least 0.3 GB larger than the old size")
            validation_result = validate_filesystem_size(new_size)
            if validation_result:
                raise ProvisionValidationException(validation_result)

    def _resize_volume(self):
        self._task.set_description("Resizing volume")
        device, volume = self._get_device_and_volume()
        self._original_size = volume.get_size()
        volume.update_size(self._proper_size)
        self._volume = volume

    def _get_steps(self):
        res = [
            self._resize_volume,
            self._rescan]
        datastore = get_datastore_by_serial_number(self._esx_host, self._proper_serial)
        if datastore is not None:
            res.extend([self._resize_datastore])
        res.extend([
            self._rescan_other_hosts,
            self._finish])
        return res


class Resize(PluginResource):
    @json_response
    def resize_validation(self, moref, device_name, size, **kwargs):
        host = moref_to_object(moref)
        proper_serial = fix_serial(device_name)
        proper_size = convert_size(size)
        return HostResizeVolumeTask(host, proper_serial, proper_size).assert_resize_available()

    @json_response
    def get_resize_tasks(self, moref, device_name, **kwargs):
        host = moref_to_object(moref)
        proper_serial = fix_serial(device_name)
        proper_size = 0
        return HostResizeVolumeTask(host, proper_serial, proper_size).get_task_list()

    @json_response
    def run_resize_task(self, moref, device_name, size, **kwargs):
        # Note: device_name contains the volume serial
        host = moref_to_object(moref)
        task_key = (HostResizeVolumeTask, moref)
        proper_serial = fix_serial(device_name)
        proper_size = convert_size(size)
        task_object_creator = lambda: HostResizeVolumeTask(host, proper_serial, proper_size)
        return self.next_task_step(task_key, task_object_creator)
