import cherrypy
import logging
from infi.vendata.vmware_powertools.vcenter.shortcuts import create_vcenter_task
from infi.vendata.vmware_powertools.utils import spawn
from infi.vendata.vmware_powertools.patches import with_host
from infi.vendata.vmware_powertools.utils import get_infinisdk_volume_by_device, get_infinidat_device_by_serial

logger = logging.getLogger(__name__)


def update_device_name(esx_host, device):
    """ Change the device name to the volume name if they are different. Returns True if a change was made. """
    actual = device.get_display_name()
    expected = getattr(device.get_vendor(), 'get_volume_name', device.get_vendor().get_system_name)()
    if actual == expected:
        return False
    uuid = device.get_uuid()
    logger.info("Updating display name for {!r} with uuid {!r} to {!r}".format(device, uuid, expected))
    esx_host.configManager.storageSystem.UpdateScsiLunDisplayName(uuid, expected)
    return True

def get_datastore_by_serial_number(host, serial_number):
    from infi.vendata.vmware_powertools.vcenter.shortcuts import iter_infinidat_datastores
    with with_host(host):
        for datastore, infinidat_devices in iter_infinidat_datastores(host):
            for device in infinidat_devices:
                if device.get_scsi_serial_number() == serial_number:
                    return datastore
    return None

def create_vmfs_datastore(esx_host, block_device, volume):
    # for easy mock patching
    datastore_system = esx_host.configManager.datastoreSystem
    device_path = block_device.get_block_access_path()
    create_options = datastore_system.QueryVmfsDatastoreCreateOptions(device_path, 5)[0]
    create_options.spec.vmfs.volumeName = volume.get_name()
    return datastore_system.CreateVmfsDatastore(create_options.spec)

def create_vmfs_datastore_with_metadata(esx_host, block_device, volume):
    with with_host(esx_host):
        new_datastore = create_vmfs_datastore(esx_host, block_device, volume)
        volume.set_metadata_from_dict(dict(filesystem_type="VMFS", mount_point=new_datastore.host[0].mountInfo.path))
        return new_datastore

def create_vmfs_datastore_by_serial(esx_host, proper_serial):
    from infi.vendata.vmware_powertools.vcenter.unmap import VolumeUnmapBase
    VolumeUnmapBase(esx_host, proper_serial).assert_not_being_used_as_rdm()
    device = get_infinidat_device_by_serial(esx_host, proper_serial)
    with with_host(esx_host):
        volume = get_infinisdk_volume_by_device(device)
        new_datastore = create_vmfs_datastore_with_metadata(esx_host, device, volume)
        datastore_changed(proper_serial)
        return new_datastore

def mount_vmfs_datastore(esx_host, block_device):
    # A newly discovered vmfs volume will be automatically mounted
    # (unless it has been explicitly unmounted before - but this function is called during provisioning / create
    # filesystem, so we're safe from that)
    # We leave this here for mock patching
    pass

def mount_vmfs_datastore_by_serial(esx_host, proper_serial):
    device = get_infinidat_device_by_serial(esx_host, proper_serial)
    mount_vmfs_datastore(esx_host, device)

def host_datastores_changed(moref, system_serial):
    # Called when a datastore was added to the host or removed from it.
    # Clear the cache of the host's datastores
    host_datastores = cherrypy.tree.apps[''].root.plugin.views.host.datastores
    host_datastores.clear_cached_entry(host_datastores.get_row_data, moref, system_serial)

def datastore_changed(serial):
    # Clear the caches of the hosts that are mapped to the datastore
    # Should be called under with_host
    host_datastores = cherrypy.tree.apps[''].root.plugin.views.host.datastores
    host_datastores.clear_cache_by_datastore(serial)

def fix_serial(serial):
    return str(serial.replace(u'\ufffd', '\x00'))

def get_system_serial(volume_serial):
    """
    Extract the system serial from the hexadecimal volume serial.
    """
    return int(volume_serial[11:11+4], 16)


class TaskMixin(object):
    # For some tasks like unmap and rename we use the function 'step' from pyvmomi_wrapper's Task object, which
    # runs all the steps that are needed to finish the task one after the other. But for other tasks like
    # provisioning and rename we want to have progress indication on the webpage (with actions_modal), and for that
    # we use this class.
    # 'start' spawns '_run' which spawns a greenlet for each step. This acheives the same thing as Task.step,
    # except we have a greenlet that we can monitor for each step. 'next' is called by the client, and it only
    # waits for the next greenlet to end, so that the client's browser can get progress indication.

    # subclasses need to implement _get_steps and define the members _esx_object and _task_name for the VMware task
    def _run(self, steps):
        percentage = 0
        increment = 100 / len(steps)
        with self._task:
            for step in steps:
                greenlet = spawn(step)
                self._queue.put(greenlet)
                greenlet.get()      # join + raise exception if there was any
                percentage += increment
                self._task.update_progress(percentage)
            self._task.update_progress(100)

    def start(self):
        from gevent.queue import Queue
        self._task = create_vcenter_task(self._esx_object, self._task_name)
        self._queue = Queue()
        return spawn(self._run, self._get_steps())

    def next(self):
        greenlet = self._queue.get()
        return greenlet.get()      # join + raise exception if there was any

    def _finish(self):
        return StopIteration       # we don't raise to avoid logging; 'next_task_step' checks return value for this
