from pyVmomi import vim
from infi.vendata.vmware_powertools.vcenter.shortcuts import get_cluster_datastores


def _get_children(obj, motype):
    if isinstance(obj, vim.Datacenter):
        if motype == vim.HostSystem:
            return obj.hostFolder.childEntity
        if motype == vim.Folder:
            return [obj.vmFolder]
    if isinstance(obj, vim.ComputeResource) and motype == vim.Datastore:
        return get_cluster_datastores(obj)
    if isinstance(obj, vim.ComputeResource):
        return obj.host + obj.resourcePool.resourcePool
    if isinstance(obj, vim.ResourcePool) and motype == vim.Datastore:
        return get_cluster_datastores(obj.owner)
    if isinstance(obj, vim.ResourcePool):
        return obj.resourcePool
    if isinstance(obj, vim.HostSystem) and motype == vim.Datastore:
        return obj.datastore
    if isinstance(obj, vim.Folder):
        return [child for child in obj.childEntity if isinstance(child, (vim.Folder, vim.Datacenter, motype))]
    return []


def get_icon_name(obj):
    if isinstance(obj, (vim.Datacenter, vim.VirtualMachine, vim.Datastore, vim.ComputeResource, vim.HostSystem, vim.ResourcePool)):
        clsname = obj.__class__.__name__[4:].replace("ClusterComputeResource", "ComputeResource")
        return "/static/vmware/images/fancytree/{}.png".format(clsname)


def get_free_space_for_datastore(child):
    from capacity import TiB, GiB, MiB, KiB, byte
    if isinstance(child, vim.Datastore):
        free_space = float(child.summary.freeSpace) * byte
        # we don't use str(free_space) becasue capacity str decides whether to use XiB or XB units
        # by itself. We only want the XiB units (and call them XB) because that's the terminology in VMware
        for unit in [TiB, GiB, MiB]:
            if free_space >= unit:
                return "{:.2f} {}B".format(free_space / unit, str(unit)[2])
        return "{:.2f} KB".format(free_space / KiB)


def treeview_data(client, moref, motype_name, toplevel=False):
    obj = client.get_managed_object_by_reference(moref)
    motype = getattr(vim, motype_name)
    active_motype = motype if motype != vim.HostSystem else (vim.HostSystem, vim.ResourcePool, vim.ClusterComputeResource)
    children = _get_children(obj, motype)
    res = []
    for child in children:
        if isinstance(child, vim.ComputeResource) and not isinstance(child, vim.ClusterComputeResource):
            # single hosts (that are not in a cluster) have a ComputeResource above them, that we don't need to show.
            # From http://pubs.vmware.com/vi30/sdk/ReferenceGuide/vim.ComputeResource.html:
            # "If the compute resource is a standalone type, then this list contains just one element."
            child = child.host[0]
        child_info = dict(title=child.name,
                          lazy=len(_get_children(child, motype)) > 0,
                          folder=isinstance(child, vim.Folder),
                          data=dict(moref=client.get_reference_to_managed_object(child),
                                    icon=get_icon_name(child),
                                    activatable=isinstance(child, active_motype)),
                                    freespace=get_free_space_for_datastore(child))
        res.append(child_info)
    if toplevel:
        res = [dict(title=obj.name,
                    expanded=True,
                    folder=isinstance(obj, vim.Folder),
                    children=res,
                    data=dict(activatable=False,
                              icon=get_icon_name(obj)))]
    return res
